from typing import Any, Optional

from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from pydantic import BaseModel, validator
import scipy.stats
from torch import Tensor

from camfi.datamodel.via import ViaRegionAttributes
from camfi.wingbeat import WingbeatSuppFigPlotter, BcesResult, WeightedGaussian


class MatplotlibWingbeatSuppFigPlotter(WingbeatSuppFigPlotter):
    """Implementation of WingbeatSuppFigPlotter using matplotlib."""

    def __call__(
        self,
        region_attributes: ViaRegionAttributes,
        region_of_interest: Tensor,
        mean_autocorrelation: Tensor,
    ) -> None:
        """Plot a supplementary figure of wingbeat extraction, saving the image to a
        file.

        Parameters
        ----------
        region_attributes : ViaRegionAttributes
            With fields calculated (e.g. by WingbeatExtractor.process_blur).
        region_of_interest : Tensor
            Greyscale image Tensor displaying region of interest.
        mean_autocorrelation : Tensor
            1-d Tensor with values containing autocorrrelation along axis 1 of
            region_of_interest.
        """

        fig = plt.figure()

        ax1 = fig.add_subplot(
            211,
            title=f"Linearised view of insect motion blur (moth {self.annotation_idx})",
        )
        ax1.imshow(region_of_interest.numpy())

        if region_attributes.snr is not None:
            ax2_title = (
                f"Autocorrelation along motion blur (SNR: {region_attributes.snr:.2f})"
            )
        else:
            ax2_title = f"Autocorrelation along motion blur (No peak found)"

        ax2 = fig.add_subplot(
            212,
            title=ax2_title,
            ylabel="Correlation",
            xlabel="Distance (pixel columns)",
        )
        ax2.plot(mean_autocorrelation.numpy())

        if region_attributes.best_peak is not None:
            best_peak = region_attributes.best_peak
            ax1.axvline(best_peak, c="r")
            ax2.axvline(best_peak, c="r")
            ax2.axvspan(
                best_peak // 4, (best_peak * 3) // 4, color="k", alpha=0.25, zorder=0
            )
            ax2.axvspan(
                (best_peak * 5) // 4,
                (best_peak * 7) // 4,
                color="k",
                alpha=0.25,
                zorder=0,
            )

            try:
                ymin = float(mean_autocorrelation.min())
                ymax = float(mean_autocorrelation[best_peak])
                yrange = ymin - ymax
                ymin -= yrange * 0.05
                ymax += yrange * 0.05
                ax2.set_ylim(ymin, ymax)
            except ValueError:
                pass

        filepath = self.get_filepath()
        try:
            fig.savefig(str(filepath))
        except FileNotFoundError:
            filepath.parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(filepath)

        plt.close(fig)


def plot_herror_bars(
    axes: plt.Axes, lower: np.ndarray, upper: np.ndarray, y: np.ndarray, **kwargs
) -> list[plt.Line2D]:
    """Plots horizontal error bars on a set of axes.

    Parameters
    ----------
    axes : plt.Axes
        Axes to plot error bars on.
    lower : np.ndarray
        Lower x-values of error bars.
    upper : np.ndarray
        Upper x-values of error bars. Should have same shape as lower.
    y : np.ndarray
        y-values. Should have same shape as upper and lower.
    **kwargs : dict
        Passed to plt.plot.
    """
    assert (
        lower.shape == upper.shape == y.shape
    ), f"Shapes do not match. Got {lower.shape}, {upper.shape}, {y.shape}."
    return axes.plot(
        np.stack((lower, upper)), np.broadcast_to(y, (2,) + y.shape), **kwargs
    )


class MatplotlibWingbeatFrequencyPlotter(BaseModel):
    """Produces plots of wingbeat frequency measurements. Most parameters are just to
    control exactly how the plot looks, and the defaults should make it look pretty.
    Probably only need to set polyline_regions, snr_thresh, class_mask, gmm_results, and
    bces_results parameters, but everything else is exposed for fine control.

    A few fields come up in the class signature, but should not be set manually. These
    are ommitted from the parameter list.

    Parameters
    ----------
    polyline_regions : pd.DataFrame
        Dataframe containing wingbeat data. Each entry should be a polyline region with
        all columns filled. (See camfi.datamodel.via.ViaProject.to_region_dataframe for
        dataframe creation).
    snr_thresh : float
        SNR threshold to plot.
    class_mask : Optional[np.ndarray]
        Array of class indices. Should have same length as the thresholded dataset,
        obtained by calling polyline_regions[polyline_regions["snr"] >= snr_thresh.
    gmm_results : Optional[list[WeightedGaussian]]
        If provided, a log10 Gaussian mixture model will be plotted on the horizontal
        marginal distribution histogram of the snr_vs_pwf plot. See camfi.wingbeat.GMM
        for creating this list.
    bces_results : Optional[list[BcesResult]]
        If provided, BCES regression lines will be plotted on the l_vs_pdt plot. See
        camfi.wingbeat.BcesEM for creating this list.
    figsize : tuple[float, float]
        Size (width, height) of figure in inches (*shudders* ...This is the units
        matplotlib uses).
    left_border : float
        Proportion of figure width to pad on left side.
    bottom_border : float
        Proportion of figure height to pad on bottom side.
    snr_vs_pwf_ax_width : float
        Width of snr_vs_pwf plot as a proportion of total figure width.
    snr_vs_pwf_ax_height : float
        Height of snr_vs_pwf plot as a proportion of total figure height.
    hist_height : float
        Height of snr_vs_pwf marginal histograms as a proportion of total figure height
        and width.
    l_vs_pdt_spacing : float
        Horizontal spacing between l_vs_pdt plot and snr_vs_pwf plot. Negative values
        indicate that plots should overlap.
    snr_vs_pwf_alpha : float
        Alpha level (opacity) to apply to error bars and histograms for snr_vs_pwf plot.
    snr_vs_pwf_abovethresh_c : str
        Colour for above-threshold error bars on snr_vs_pwf plot.
    snr_vs_pwf_belowthresh_c : str
        Colour for below-threshold error bars on snr_vs_pwf plot.
    snr_thresh_line_c : str
        Colour of SNR threshold line.
    errorbar_lw : float
        Line width of error bars. Applied across both subfigures.
    gmm_plot_range_stdevs : float
        Number of standard deviations either side of means to plot for GMM pdfs.
    gmm_lw : float
        Line width for GMM pdfs. By default this is bold for easier viewing.
    l_vs_pdt_alpha : float
        Alpha level (opacity) to apply to error bars for l_vs_pdt plot.
    class_colours : list[str]
        list of colours to use for plotting multiple classes (both for GMM and BCES,
        and class_mask). This will apply the colours to the classes in the order they
        are given. If class_mask is not supplied, the last colour in the list is used
        for all datapoints.
    snr_vs_pwf_title : str
        Title for snr_vs_pwf plot.
    snr_vs_pwf_title_y : float = 0.87
        Vertical position of title for snr_vs_pwf plot.
    l_vs_pdt_title : str = " (b)"
        Title for l_vs_pdt plot.
    l_vs_pdt_title_y : float = 0.91
        Vertical position of title for l_vs_pdt plot.
    title_font_dict : dict[str, Any] = {"fontweight" : "bold"}
        Defines font parameters for subfig titles.
    """

    polyline_regions: pd.DataFrame
    snr_thresh: float = 4.0
    class_mask: np.ndarray = None  # type: ignore[assignment]
    gmm_results: Optional[list[WeightedGaussian]] = None
    bces_results: Optional[list[BcesResult]] = None
    figsize: tuple[float, float] = (7.5, 5.2)
    left_border: float = 0.1
    bottom_border: float = 0.1
    snr_vs_pwf_ax_width: float = 0.4
    snr_vs_pwf_ax_height: float = 0.4
    hist_height: float = 0.2
    l_vs_pdt_spacing: float = -0.12
    snr_vs_pwf_alpha: float = 1.0
    snr_vs_pwf_abovethresh_c: str = "k"
    snr_vs_pwf_belowthresh_c: str = "grey"
    snr_thresh_line_c: str = "r"
    errorbar_lw: float = 1
    gmm_plot_range_stdevs: float = 4.0
    gmm_lw: float = 3
    l_vs_pdt_alpha: float = 0.5
    class_colours: list[str] = [
        "tab:blue",
        "tab:green",
        "tab:orange",
        "tab:red",
        "tab:purple",
        "k",
    ]
    snr_vs_pwf_title: str = " (a)"
    snr_vs_pwf_title_y: float = 0.87
    l_vs_pdt_title: str = " (b)"
    l_vs_pdt_title_y: float = 0.91
    title_font_dict: dict[str, Any] = {"fontweight": "bold"}
    fig: plt.Figure = None  # type: ignore[assignment]
    snr_vs_pwf_ax: plt.Axes = None  # type: ignore[assignment]
    histx_ax: plt.Axes = None  # type: ignore[assignment]
    histy_ax: plt.Axes = None  # type: ignore[assignment]
    l_vs_pdt_ax: plt.Axes = None  # type: ignore[assignment]
    above_thresh: pd.DataFrame = None  # type: ignore[assignment]
    below_thresh: pd.DataFrame = None  # type: ignore[assignment]

    class Config:
        arbitrary_types_allowed = True

    @validator("class_mask", pre=True, always=True)
    def class_mask_same_length_as_abovethresh(cls, v, values):
        n_abovethresh = np.count_nonzero(
            values["polyline_regions"]["snr"] >= values["snr_thresh"]
        )
        if v is None:
            v = np.zeros((n_abovethresh,), dtype="i4") - 1
        assert (
            len(v) == n_abovethresh
        ), "class_mask must have one value for each above-snr-thresh datapoint."
        return v

    def _init_figure(self) -> None:
        """Initialises Figure and all Axes into self.fig."""
        self.fig = plt.figure(figsize=self.figsize)
        self._init_snr_vs_pwf_ax()
        self._init_histx_ax()
        self._init_histy_ax()
        self._init_l_vs_pdt_ax()

    def _init_snr_vs_pwf_ax(self) -> None:
        """Initialises Axes into self.snr_vs_pwf_ax."""
        self.snr_vs_pwf_ax = self.fig.add_axes(
            [
                self.left_border,
                self.bottom_border,
                self.snr_vs_pwf_ax_width,
                self.snr_vs_pwf_ax_height,
            ],
            xlabel="Preliminary wingbeat frequency (Hz)",
            ylabel="SNR",
            xscale="log",
        )

    def _init_histx_ax(self) -> None:
        """Initialises Axes into self.histx_ax."""
        self.histx_ax = self.fig.add_axes(
            [
                self.left_border,
                self.bottom_border + self.snr_vs_pwf_ax_height,
                self.snr_vs_pwf_ax_width,
                self.hist_height,
            ],
            sharex=self.snr_vs_pwf_ax,
        )
        self.histx_ax.axis("off")

    def _init_histy_ax(self) -> None:
        """Initialises Axes into self.histy_ax."""
        self.histy_ax = self.fig.add_axes(
            [
                self.left_border + self.snr_vs_pwf_ax_width,
                self.bottom_border,
                self.hist_height,
                self.snr_vs_pwf_ax_height,
            ],
            sharey=self.snr_vs_pwf_ax,
        )
        self.histy_ax.axis("off")

    def _init_l_vs_pdt_ax(self) -> None:
        """Initialises Axes into self.l_vs_pdt_ax."""
        self.l_vs_pdt_ax = self.fig.add_axes(
            [
                self.left_border + self.snr_vs_pwf_ax_width + self.l_vs_pdt_spacing,
                self.bottom_border + self.snr_vs_pwf_ax_height + self.l_vs_pdt_spacing,
                1.0
                - self.left_border
                - self.snr_vs_pwf_ax_width
                - self.l_vs_pdt_spacing,
                1.0
                - self.bottom_border
                - self.snr_vs_pwf_ax_height
                - self.l_vs_pdt_spacing,
            ],
            ylabel="$L$ (pixels)",
            xlabel="$P∆t$ (pixels · s)",
        )

    def _apply_snr_thresh(self) -> None:
        """Splits self.polyline_regions into self.above_thresh and self.below_thresh."""
        above_thresh_mask = self.polyline_regions["snr"] >= self.snr_thresh
        self.above_thresh = self.polyline_regions[above_thresh_mask]
        self.below_thresh = self.polyline_regions[~above_thresh_mask]

    def _plot_snr_vs_pwf(self) -> None:
        """Plots scatter plot of SNR vs. Preliminary wingbeat frequency."""
        # Plot above-threshold data
        plot_herror_bars(
            self.snr_vs_pwf_ax,
            self.above_thresh["wb_freq_down"],
            self.above_thresh["wb_freq_up"],
            self.above_thresh["snr"],
            c=self.snr_vs_pwf_abovethresh_c,
            alpha=self.snr_vs_pwf_alpha,
            lw=self.errorbar_lw,
        )
        # Plot below-threshold data
        plot_herror_bars(
            self.snr_vs_pwf_ax,
            self.below_thresh["wb_freq_down"],
            self.below_thresh["wb_freq_up"],
            self.below_thresh["snr"],
            c=self.snr_vs_pwf_belowthresh_c,
            alpha=self.snr_vs_pwf_alpha,
            lw=self.errorbar_lw,
        )

        self.snr_vs_pwf_ax.axhline(
            self.snr_thresh, c=self.snr_thresh_line_c, zorder=0, label="SNR Threshold"
        )

    def _plot_marginal_hists(self) -> None:
        """Plots marginal histograms for SNR vs. Preliminary wingbeat frequency."""
        # Horizontal marginal
        hx, bx, p = self.histx_ax.hist(
            np.concatenate(
                [
                    self.polyline_regions["wb_freq_down"],
                    self.polyline_regions["wb_freq_up"],
                ]
            ),
            bins=np.logspace(
                np.log10(min(self.polyline_regions["wb_freq_down"])),
                np.log10(max(self.polyline_regions["wb_freq_up"])),
            ),
            facecolor=self.snr_vs_pwf_belowthresh_c,
            alpha=self.snr_vs_pwf_alpha,
        )

        hx_filt, bx_filt, p = self.histx_ax.hist(
            np.concatenate(
                [self.above_thresh["wb_freq_down"], self.above_thresh["wb_freq_up"]]
            ),
            bins=bx,
            facecolor=self.snr_vs_pwf_abovethresh_c,
            alpha=self.snr_vs_pwf_alpha,
        )

        # Plot GMM
        if self.gmm_results is not None:
            scaling = np.mean(hx_filt * (bx_filt[1:] - bx_filt[:-1])) / 2
            for class_i in range(len(self.gmm_results)):
                pdf_x = np.logspace(
                    self.gmm_results[class_i].mean
                    - self.gmm_results[class_i].std * self.gmm_plot_range_stdevs,
                    self.gmm_results[class_i].mean
                    + self.gmm_results[class_i].std * self.gmm_plot_range_stdevs,
                    num=100,
                )
                self.histx_ax.plot(
                    pdf_x,
                    scaling
                    * self.gmm_results[class_i].weight
                    * scipy.stats.norm.pdf(
                        np.log10(pdf_x),
                        loc=self.gmm_results[class_i].mean,
                        scale=self.gmm_results[class_i].std,
                    ),
                    c=self.class_colours[class_i],
                    linewidth=self.gmm_lw,
                )

        # Vertial marginal
        # First need to pin bin edges to snr_thresh to avoid overlap
        min_snr = self.polyline_regions["snr"].min()
        max_snr = self.polyline_regions["snr"].max()
        nbins = 50
        by = np.linspace(
            min_snr - (max_snr - min_snr) / nbins,
            max_snr,
            num=nbins + 1,
        )
        by += self.snr_thresh - by[by <= self.snr_thresh][-1]

        hy, by, p = self.histy_ax.hist(
            self.polyline_regions["snr"],
            bins=by,
            orientation="horizontal",
            facecolor=self.snr_vs_pwf_belowthresh_c,
            alpha=self.snr_vs_pwf_alpha,
        )

        self.histy_ax.hist(
            self.above_thresh["snr"],
            bins=by,
            orientation="horizontal",
            facecolor=self.snr_vs_pwf_abovethresh_c,
            alpha=self.snr_vs_pwf_alpha,
        )

        # SNR threshold line should be continued into the marginal
        self.histy_ax.axhline(
            self.snr_thresh, c=self.snr_thresh_line_c, zorder=1, label="SNR Threshold"
        )

    def _plot_l_vs_pdt(self) -> None:
        """Plot blur length vs. pixel-period * ∆t for above thresh data only."""
        for class_i in np.unique(self.class_mask):
            mask = self.class_mask == class_i
            plot_herror_bars(
                self.l_vs_pdt_ax,
                self.above_thresh["best_peak"][mask] * self.above_thresh["et_up"][mask],
                self.above_thresh["best_peak"][mask] * self.above_thresh["et_dn"][mask],
                self.above_thresh["blur_length"][mask],
                c=self.class_colours[class_i],
                alpha=self.l_vs_pdt_alpha,
                lw=self.errorbar_lw,
            )

    def _plot_l_vs_pdt_regressions(self) -> None:
        """Plots regression lines."""
        if self.bces_results is not None:
            for i in range(len(self.bces_results)):
                xmax = (
                    self.above_thresh["best_peak"] * self.above_thresh["et_dn"]
                ).max()
                self.l_vs_pdt_ax.plot(
                    [0, xmax],
                    [
                        self.bces_results[i].y_intercept,
                        self.bces_results[i].y_intercept
                        + self.bces_results[i].gradient * xmax,
                    ],
                    c=self.class_colours[i],
                )

    def _add_titles(self) -> None:
        """Adds titles to subfigures."""
        title_y = 0.88
        a_title = self.snr_vs_pwf_ax.set_title(
            self.snr_vs_pwf_title,
            fontdict=self.title_font_dict,
            loc="left",
            y=self.snr_vs_pwf_title_y,
        )
        b_title = self.l_vs_pdt_ax.set_title(
            self.l_vs_pdt_title,
            fontdict=self.title_font_dict,
            loc="left",
            y=self.l_vs_pdt_title_y,
        )

    def plot(self) -> plt.Figure:
        """Produces plots."""
        # Initialise axes
        self._init_figure()

        # Apply snr threshold
        self._apply_snr_thresh()

        # Plot
        self._plot_snr_vs_pwf()
        self._plot_marginal_hists()
        self._plot_l_vs_pdt()
        self._plot_l_vs_pdt_regressions()
        self._add_titles()

        return self.fig


def plot_activity_levels_summary(
    df: pd.DataFrame,
    ax: plt.Axes,
    x_column: str = "daynumber",
    bin_width: float = 10 / 1440,
    **kwargs,
) -> list[plt.Line2D]:
    """Produces a histogram plot of df["n_annotations"] vs. df[x_column].

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame which must contain columns "n_annotations" and x_column.
    ax : plt.Axes
        Matplotlib axes to plot histogram on.
    x_column : str
        Column of df to use as x-axis data.
    bin_width : float
        Width of histogram bins, in same units as x_column.
    **kwargs
        Passed to plt.Axes.plot.

    Returns
    -------
    lines : list[plt.Line2D]
        list of plot lines.
    """
    h, bx, by = np.histogram2d(
        df[x_column],
        df["n_annotations"],
        bins=[
            np.arange(
                min(df[x_column]) - bin_width / 2,
                max(df[x_column]) + bin_width,
                bin_width,
            ),
            np.arange(-0.5, max(df["n_annotations"]) + 1, 1),
        ],
    )
    b_midpoints = (bx[:-1] + bx[1:]) / 2
    n_annotations_perbin = (h * np.arange(h.shape[1]).reshape((1, h.shape[1]))).sum(
        axis=1
    )
    n_annotations_perbin[h.sum(axis=1) == 0.0] = np.nan

    return ax.plot(b_midpoints, n_annotations_perbin, drawstyle="steps-mid", **kwargs)


def plot_activity_levels_summaries(
    df: pd.DataFrame,
    locations: list[str],
    sub_figsize: tuple[float, float] = (9, 5),
    ax_kwargs: dict = {},
    separate_plots: bool = True,
    **kwargs,
) -> plt.Figure:
    """Calls plot_activity_levels_summary for multiple locations, to produce multiple
    histogram plots of activity levels.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame which must contain columns "n_annotations", and be indexed by location
        in locations.
    locations : list[str]
        list of locations to plot data for. [df.loc[location] for location in locations]
        should be valid.
    sub_figsize : tuple[float, float]
        Figure size (in inches). If separate_plots is True, then each subfigure will be
        this size.
    ax_kwargs : dict
        Keyword arguments passed to plt.Figure.add_subplot.
    separate_plots : bool
        If True (default), each plot will be put on a separate subfigure. If False, they
        will all be plotted on the same axes.
    **kwargs
        Passed to plot_activity_levels_summary.

    Returns
    -------
    fig : plt.Figure
        Figure containing plots.
    """
    fig = plt.figure(figsize=(sub_figsize[0], sub_figsize[1] * len(locations)))
    sharex: Optional[plt.Axes] = None
    if separate_plots is False:
        ax = fig.add_subplot(111, **ax_kwargs)
    for i, location in enumerate(locations):
        if separate_plots is True:
            ax = fig.add_subplot(
                100 * len(locations) + 11 + i,
                title=location,
                sharex=sharex,
                **ax_kwargs,
            )
            sharex = ax
        plot_activity_levels_summary(df.loc[location], ax, label=location, **kwargs)

    return fig


def plot_daily_temperatures(
    df: pd.DataFrame,
    ax: plt.Axes,
    minimum_col: str = "temperature_minimum_evening_degC",
    **kwargs,
) -> list[plt.Line2D]:
    """Plots daily temperature values on ax.

    Parameters
    ----------
    df : pd.DataFrame
       DataFrame with columns "daynumber", "temperature_maximum_degC",
       "temperature_3pm_degC", "temperature_9am_degC", and minimum_col.
    ax : plt.Axes
        Axes to put plots on.
    minimum_col : str
       Name of column for minimum daily temperature. E.g. "temperature_minimum_degC", or
       "temperature_minimum_evening_degC" (default).
    **kwargs
       Passed to each call to ax.plot.

    Returns
    -------
    lines : list[plt.Line2D]
        list of plot lines.
    """
    lines = []
    lines.extend(
        ax.plot(
            df["daynumber"],
            df["temperature_maximum_degC"],
            c="r",
            label="Daily Max.",
            **kwargs,
        )
    )
    lines.extend(
        ax.plot(
            df["daynumber"],
            df["temperature_3pm_degC"],
            c="r",
            alpha=0.4,
            label="3pm Temp.",
            **kwargs,
        )
    )
    lines.extend(
        ax.plot(
            df["daynumber"],
            df[minimum_col],
            c="b",
            label="Daily Min.",
            **kwargs,
        )
    )
    lines.extend(
        ax.plot(
            df["daynumber"],
            df["temperature_9am_degC"],
            c="b",
            alpha=0.4,
            label="9am Temp.",
            **kwargs,
        )
    )
    return lines


def plot_maelstroms(df: pd.DataFrame, ax: plt.Axes, **kwargs) -> list[plt.Line2D]:
    """Plots "n_annotations" vs "daynumber", with data taken from df.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with columns "daynumber" and "n_annotations".
    ax : plt.Axes
        Axes to plot onto.
    **kwargs
        Passed to ax.plot.
    """
    return ax.plot(df["daynumber"], df["n_annotations"], **kwargs)


def plot_maelstroms_with_temperature(
    df: pd.DataFrame,
    ax: plt.Axes,
    maelstrom_kwargs: dict = {},
    temperatures_kwargs: dict = {},
) -> tuple[list[plt.Line2D], list[plt.Line2D]]:
    """Calls plot_maelstrom, then makes a twinx axes from ax, and calls
    plot_daily_temperatures.

    Parameters
    ----------
    df : DataFrame
        Passed to plot_maelstrom and plot_daily_temperatures.
    ax : plt.Axes
        Axes to plot maelstrom data onto. Temperature data is put on a new Axes created
        by calling ax.twinx().
    maelstrom_kwargs : dict
        Keyword arguments passed to plot_maelstrom.
    temperatures_kwargs : dict
        Keyword arguments passed to plot_daily_temperatures.

    Returns
    -------
    maelstrom_lines : list[plt.Line2D]
        Maelstrom plot lines.
    temperature_lines : list[plt.Line2D]
        Temperature plot lines.
    """
    maelstrom_lines = plot_maelstroms(df, ax, **maelstrom_kwargs)
    ax2 = ax.twinx()
    ax2.set_ylabel("Temperature (°C)")
    temperature_lines = plot_daily_temperatures(df, ax2, **temperatures_kwargs)

    return maelstrom_lines, temperature_lines



"""Defines geometry related classes and methods used throughout camfi. Depends on
camfi.util."""

from __future__ import annotations

from abc import ABC, abstractmethod
from math import atan2, degrees, sqrt
from typing import Optional

from pydantic import (
    BaseModel,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveInt,
    validator,
)
from shapely.geometry import LineString
import torch
from torchvision.transforms import InterpolationMode
from torchvision.transforms.functional import pad, rotate

from camfi.util import smallest_enclosing_circle, Field


class BoundingBox(BaseModel):
    """Defines a bounding box, and provides various convenience methods for working with
    boxes on images.

    Parameters
    ----------
    x0: NonNegativeInt
        Minimum inclusive (horizontal) x-coordinate of box.
    y0: NonNegativeInt
        Minimum inclusive (vertical) y-coordinate of box.
    x1: NonNegativeInt
        Maximum exclusive (horizontal) x-coordinate of box.
    y1: NonNegativeInt
        Maximum exclusive (vertical) y-coordinate of box.
    """

    x0: NonNegativeInt = Field(
        ..., description="Minimum inclusive (horizontal) x-coordinate of box."
    )
    y0: NonNegativeInt = Field(
        ..., description="Minimum inclusive (vertical) y-coordinate of box."
    )
    x1: NonNegativeInt = Field(
        ..., description="Maximum exclusive (horizontal) x-coordinate of box."
    )
    y1: NonNegativeInt = Field(
        ..., description="Maximum exclusive (vertical) y-coordinate of box."
    )

    class Config:
        schema_extra = {
            "description": "Defines a bounding box with integer (pixel) coordinates."
        }

    @validator("x1")
    def x1_gt_x0(cls, v, values):
        """Pydantic validation method, called when instantiating BoundingBox.
        Ensures that x1 > x0."""
        if "x0" in values and v <= values["x0"]:
            raise ValueError("x1 and y1 must be larger than x0 and y0")
        return v

    @validator("y1")
    def y1_gt_y0(cls, v, values):
        """Pydantic validation method, called when instantiating BoundingBox.
        Ensures that y1 > y0."""
        if "y0" in values and v <= values["y0"]:
            raise ValueError("x1 and y1 must be larger than x0 and y0")
        return v

    @classmethod
    def from_shape(
        self, shape: tuple[PositiveInt, PositiveInt], border: NonNegativeInt = 0
    ) -> BoundingBox:
        """Creates an instance of BoundingBox from an image shape, useful for defining
        a region of interest within an image, not too close to the edge.

        Parameters
        ----------
        shape: tuple[PositiveInt, PositiveInt]
            Shape of image (height, width).
        border: NonNegativeInt
            Width of border. If 0 (default), then the bounding box will contain the
            entire image.

        Returns
        -------
        box : BoundingBox
            Optionally contracted bounding box of image.

        Examples
        --------
        >>> BoundingBox.from_shape((10, 15))
        BoundingBox(x0=0, y0=0, x1=16, y1=11)

        >>> BoundingBox.from_shape((10, 15), border=3)
        BoundingBox(x0=3, y0=3, x1=13, y1=8)
        """
        return BoundingBox(
            x0=border, y0=border, x1=shape[1] + 1 - border, y1=shape[0] + 1 - border
        )

    @property
    def shape(self) -> tuple[int, int]:
        """Gets the (height, width) of the bounding box.

        Returns
        -------
        shape : tuple[int, int]
            Height and width of the bounding box

        Examples
        --------
        >>> BoundingBox(x0=5, y0=1, x1=15, y1=3).shape
        (2, 10)
        """
        return (self.y1 - self.y0, self.x1 - self.x0)

    def add_margin(
        self,
        margin: NonNegativeInt,
        shape: Optional[tuple[PositiveInt, PositiveInt]] = None,
    ) -> None:
        """Expands self by a fixed margin. Operates in-place.

        Parameters
        ----------
        margin: PositiveInt
            Margin to add to self.
        shape: Optional[tuple[PositiveInt, PositiveInt]] = (height, width)
            Shape of image. If set, will constrain self to image shape.
        """
        self.x0 = max(0, self.x0 - margin)
        self.y0 = max(0, self.y0 - margin)

        self.x1 += margin
        self.y1 += margin

        if shape is not None:
            self.x0 = min(shape[1], self.x0)
            self.y0 = min(shape[0], self.y0)
            self.x1 = min(shape[1], self.x1)
            self.y1 = min(shape[0], self.y1)

    def get_area(self) -> PositiveInt:
        """Get the area enclosed by self.

        Returns
        -------
        area : PositiveInt
            Area enclosed by box, in pixels ** 2.

        Examples
        --------
        >>> box = BoundingBox(x0=0, y0=1, x1=2, y1=3)
        >>> box.get_area()
        4
        """
        return (self.x1 - self.x0) * (self.y1 - self.y0)

    def in_box(self, box: BoundingBox) -> bool:
        """Returns True if self is contained in box.

        Parameters
        ----------
        box : BoundingBox
            Other box to test against.

        Returns
        -------
        is_in_box : bool
            True if self is completely contained within box.

        Examples
        --------
        >>> box0 = BoundingBox(x0=1, y0=2, x1=3, y1=4)
        >>> box1 = BoundingBox(x0=0, y0=1, x1=4, y1=5)
        >>> box0.in_box(box1)
        True
        >>> box1.in_box(box0)
        False

        A box is always in itself

        >>> box0.in_box(box0)
        True
        """
        return (
            self.x0 >= box.x0
            and self.y0 >= box.y0
            and self.x1 <= box.x1
            and self.y1 <= box.y1
        )

    def overlaps(self, box: BoundingBox) -> bool:
        """Returns True if two bounding boxes overlap, and False otherwise.

        Parameters
        ----------
        box: BoundingBox
            Another bounding box to compare to.

        Returns
        -------
        overlap : bool
            True if self and box intersect.

        Examples
        --------
        >>> box0 = BoundingBox(x0=0, y0=0, x1=1, y1=1)
        >>> box1 = BoundingBox(x0=2, y0=2, x1=3, y1=3)
        >>> box2 = BoundingBox(x0=0, y0=0, x1=2, y1=2)
        >>> box3 = BoundingBox(x0=1, y0=1, x1=3, y1=3)
        >>> box0.overlaps(box1)
        False
        >>> box2.overlaps(box3)
        True

        Overlaps can happen in either dimension:

        >>> box0 = BoundingBox(x0=0, y0=0, x1=2, y1=2)
        >>> box1 = BoundingBox(x0=0, y0=1, x1=1, y1=3)
        >>> box2 = BoundingBox(x0=1, y0=0, x1=3, y1=1)
        >>> box3 = BoundingBox(x0=0, y0=2, x1=2, y1=4)
        >>> box4 = BoundingBox(x0=2, y0=0, x1=4, y1=2)
        >>> box0.overlaps(box1)
        True
        >>> box0.overlaps(box2)
        True

        Overlaps are not inclusive of edges:

        >>> box0.overlaps(box3)
        False
        >>> box0.overlaps(box4)
        False
        """
        return (
            self.x0 < box.x1
            and self.y0 < box.y1
            and self.x1 > box.x0
            and self.y1 > box.y0
        )

    def intersection(self, box: BoundingBox) -> NonNegativeInt:
        """Get the intersectional area of two boxes.

        Parameters
        ----------
        box: BoundingBox
            Another bounding box to compare to.

        Returns
        -------
        intersectional_area : NonNegativeInt
            Area of intersection of two boxes.

        Examples
        --------
        >>> box0 = BoundingBox(x0=0, y0=0, x1=1, y1=1)
        >>> box1 = BoundingBox(x0=2, y0=2, x1=3, y1=3)
        >>> box2 = BoundingBox(x0=0, y0=0, x1=2, y1=2)
        >>> box3 = BoundingBox(x0=1, y0=1, x1=3, y1=3)
        >>> box0.intersection(box1)
        0
        >>> box2.intersection(box3)
        1

        Intersection is commutative

        >>> from itertools import product
        >>> pairs = product([box0, box1, box2, box3], repeat=2)
        >>> all(b0.intersection(b1) == b1.intersection(b0) for b0, b1 in pairs)
        True
        """
        if self.overlaps(box):
            return BoundingBox(
                x0=max(self.x0, box.x0),
                y0=max(self.y0, box.y0),
                x1=min(self.x1, box.x1),
                y1=min(self.y1, box.y1),
            ).get_area()
        return 0

    def intersection_over_union(self, box: BoundingBox) -> NonNegativeFloat:
        """Get the intersection over union of two boxes.

        Parameters
        ----------
        box: BoundingBox
            Another bounding box to compare to.

        Returns
        -------
        iou : NonNegativeFloat
            Intersection over Union of two boxes, between 0.0 and 1.0.

        Examples
        --------
        >>> box0 = BoundingBox(x0=0, y0=0, x1=1, y1=1)
        >>> box1 = BoundingBox(x0=2, y0=2, x1=3, y1=3)
        >>> box2 = BoundingBox(x0=0, y0=0, x1=2, y1=2)
        >>> box3 = BoundingBox(x0=1, y0=0, x1=4, y1=2)
        >>> box0.intersection_over_union(box1)
        0.0
        >>> box1.intersection_over_union(box2)
        0.0
        >>> box2.intersection_over_union(box3)
        0.25
        >>> box0.intersection_over_union(box2)
        0.25

        Intersection over union is commutative

        >>> from itertools import product
        >>> pairs = product([box0, box1, box2, box3], repeat=2)
        >>> all(
        ...     b0.intersection_over_union(b1) == b1.intersection_over_union(b0)
        ...     for b0, b1 in pairs
        ... )
        True
        """
        intersection = self.intersection(box)
        union = self.get_area() + box.get_area() - intersection

        return intersection / union

    def is_portrait(self) -> bool:
        """Returns True if bounding box is at least as tall as it is wide.

        Returns
        -------
        portrait : bool
            True if bounding box is at least as tall as it is wide.

        Examples
        --------
        >>> BoundingBox(x0=1, y0=0, x1=11, y1=10).is_portrait()
        True
        >>> BoundingBox(x0=0, y0=0, x1=11, y1=10).is_portrait()
        False
        """
        return self.y1 - self.y0 >= self.x1 - self.x0

    def crop_image(self, image: torch.Tensor) -> torch.Tensor:
        """Returns a view of an image cropped to the bounding box.

        Parameters
        ----------
        image : torch.Tensor
            With shape [..., height, width].

        Returns
        -------
        torch.Tensor
            with shape [..., self.y1 - self.y0, self.x1 - self.x0], assuming
            height <= self.y1 - self.y0 and width <= self.x1 - self.x0.

        Examples
        --------
        >>> box = BoundingBox(x0=7, x1=15, y0=3, y1=7)
        >>> grey_image = torch.zeros(10, 20)
        >>> box.crop_image(grey_image).shape
        torch.Size([4, 8])
        >>> colour_image = torch.zeros(3, 10, 20)
        >>> box.crop_image(colour_image).shape
        torch.Size([3, 4, 8])

        If BoundingBox goes outside image.shape, then output size will be truncated in
        the expected way

        >>> box = BoundingBox(x0=15, x1=25, y0=3, y1=7)
        >>> box.crop_image(grey_image).shape
        torch.Size([4, 5])
        >>> box = BoundingBox(x0=7, x1=15, y0=7, y1=13)
        >>> box.crop_image(grey_image).shape
        torch.Size([3, 8])
        """
        return image[..., self.y0 : self.y1, self.x0 : self.x1]


class ViaShapeAttributes(BaseModel, ABC):
    """Abstract base class for via region shapes. These define the geometry data of
    annotations of flying insects in images.

    Parameters
    ----------
    name : str
        Name of shape type (e.g. "point", "circle", or "polyline").
    """

    name: str

    @abstractmethod
    def get_bounding_box(self) -> BoundingBox:
        """Returns a BoundingBox object which contains the coordinates in self.
        Note that CircleShapeAttributes are treated like PointShapeAttributes (i.e. r is
        ignored).

        Returns
        -------
        box : BoundingBox
            Bounding box of ViaShapeAttributes instance.
        """

    @abstractmethod
    def in_box(self, box: BoundingBox) -> bool:
        """Returns True if all points in self are within bounding box.

        Parameters
        ----------
        box : BoundingBox
            Bounding box to test against.

        Returns
        -------
        is_in_box : bool
            True if within box.
        """

    def y_diff(self) -> PositiveInt:
        """Returns the total height (y-dimension) of the annotation (in pixels).

        Returns
        -------
        y_diff : PositiveInt
            Height of self in pixels.
        """
        bounding_box = self.get_bounding_box()
        return bounding_box.y1 - bounding_box.y0 - 1

    def intersection_over_union(self, other: ViaShapeAttributes) -> NonNegativeFloat:
        """Get the intersection over union of bounding boxes.

        Parameters
        ----------
        other: ViaShapeAttributes
            Other shape to compare to.

        Returns
        -------
        iou : NonNegativeFloat
            Intersection over Union of two bounding boxes, between 0.0 and 1.0.

        Examples
        --------
        >>> class MockShapeAttributes(ViaShapeAttributes):
        ...     bounding_box: BoundingBox
        ...     def get_bounding_box(self) -> BoundingBox:
        ...         return self.bounding_box
        ...     def in_box(self, box: BoundingBox) -> bool:
        ...         return self.bounding_box.in_box(box)
        >>> shape_attributes0 = MockShapeAttributes(
        ...     bounding_box=BoundingBox(x0=0, y0=0, x1=2, y1=1),
        ...     name="mock_shape",
        ... )
        >>> shape_attributes1 = MockShapeAttributes(
        ...     bounding_box=BoundingBox(x0=1, y0=0, x1=4, y1=1),
        ...     name="mock_shape",
        ... )
        >>> shape_attributes0.intersection_over_union(shape_attributes1)
        0.25
        """
        return self.get_bounding_box().intersection_over_union(other.get_bounding_box())


class PointShapeAttributes(ViaShapeAttributes):
    """Defines a point geometry.

    Parameters
    ----------
    cx : NonNegativeFloat
        x-coordinate of point.
    cy : NonNegativeFloat
        y-coordinate of point.
    name : str
        Name of shape (must be "point" for PointShapeAttributes instances).
    """

    cx: NonNegativeFloat
    cy: NonNegativeFloat
    name: str = Field("point", regex=r"^point$")

    def get_bounding_box(self) -> BoundingBox:
        """Finds the bounding box of the point.

        Returns
        -------
        box : BoundingBox
            Bounding box of point

        Examples
        --------
        >>> point = PointShapeAttributes(cx=10, cy=15)
        >>> point.get_bounding_box()
        BoundingBox(x0=10, y0=15, x1=11, y1=16)
        """
        return BoundingBox(
            x0=int(self.cx), y0=int(self.cy), x1=int(self.cx) + 1, y1=int(self.cy) + 1
        )

    def in_box(self, box: BoundingBox) -> bool:
        """Returns True if all points in self are within bounding box.

        Parameters
        ----------
        box: BoundingBox
           Box to test against..

        Returns
        -------
        is_in_box : bool
            True if point is within box.

        Examples
        --------
        >>> point = PointShapeAttributes(cx=2, cy=13)
        >>> point.in_box(BoundingBox(x0=2, y0=13, x1=4, y1=15))
        True
        >>> point.in_box(BoundingBox(x0=3, y0=13, x1=4, y1=15))
        False
        >>> point.in_box(BoundingBox(x0=2, y0=14, x1=4, y1=15))
        False
        >>> point.in_box(BoundingBox(x0=1, y0=13, x1=2, y1=15))
        False
        >>> point.in_box(BoundingBox(x0=2, y0=12, x1=4, y1=13))
        False
        """
        return self.get_bounding_box().in_box(box)


class CircleShapeAttributes(ViaShapeAttributes):
    """Defines a circle geometry.

    Parameters
    ----------
    cx : NonNegativeInt
        x-coordinate of centre of circle.
    cy : NonNegativeInt
        y-coordinate of centre of circle.
    name : str
        Name of shape (must be "circle" for CircleShapeAttributes instances).
    r : NonNegativeFloat
        Radius of circle.
    """

    cx: NonNegativeFloat
    cy: NonNegativeFloat
    name: str = Field("circle", regex=r"^circle$")
    r: NonNegativeFloat

    @classmethod
    def from_bounding_box(cls, box: BoundingBox) -> CircleShapeAttributes:
        """Takes a BoundingBox and returns a CircleShapeAttributes, which encloses the
        BoundingBox.

        Parameters
        ----------
        box : BoundingBox
            Box to enclose with circle

        Returns
        -------
        circle : CircleShapeAttributes
            Smallest circle enclosing box.

        Examples
        --------
        >>> from pytest import approx
        >>> box = BoundingBox(x0=0, y0=0, x1=2, y1=2)
        >>> circle = CircleShapeAttributes.from_bounding_box(box)
        >>> circle.cx == approx(1.0)
        True
        >>> circle.cy == approx(1.0)
        True
        >>> circle.r == approx(sqrt(2))
        True
        """
        all_points_x = [box.x0, box.x0, box.x1, box.x1]
        all_points_y = [box.y0, box.y1, box.y0, box.y1]
        cx, cy, r = smallest_enclosing_circle(zip(all_points_x, all_points_y))
        return CircleShapeAttributes(cx=cx, cy=cy, r=r)

    def as_point(self):
        """Converts a CircleShapeAttributes instance to a PointShapeAttributes instance.

        Returns
        -------
        point : PointShapeAttributes
            Point at centre of circle.
        """
        return PointShapeAttributes(cx=self.cx, cy=self.cy)

    def get_bounding_box(self) -> BoundingBox:
        """Finds the bounding box of the point at the centre of the circle.

        **Note: this does not return the bounding box of the entire circle, just it's
        centre.**

        Returns
        -------
        box : BoundingBox
            Bounding box of point at the centre of the circle.

        Examples
        --------
        >>> circle = CircleShapeAttributes(cx=10, cy=15, r=10)
        >>> circle.get_bounding_box()
        BoundingBox(x0=10, y0=15, x1=11, y1=16)
        """
        return self.as_point().get_bounding_box()

    def in_box(self, box: BoundingBox) -> bool:
        """Returns True if all points in self are within bounding box.

        Parameters
        ----------
        box : BoundingBox
            Box to test against.

        Returns
        -------
        is_in_box : bool
            True if centre of circle is contained in box.

        Examples
        --------
        >>> circle = CircleShapeAttributes(cx=2, cy=13, r=10)
        >>> circle.in_box(BoundingBox(x0=2, y0=13, x1=4, y1=15))
        True
        >>> circle.in_box(BoundingBox(x0=3, y0=13, x1=4, y1=15))
        False
        >>> circle.in_box(BoundingBox(x0=2, y0=14, x1=4, y1=15))
        False
        >>> circle.in_box(BoundingBox(x0=1, y0=13, x1=2, y1=15))
        False
        >>> circle.in_box(BoundingBox(x0=2, y0=12, x1=4, y1=13))
        False
        """
        return self.get_bounding_box().in_box(box)


class PolylineShapeAttributes(ViaShapeAttributes):
    """Defines a polyline geometry.

    Parameters
    ----------
    all_points_x : list[NonNegativeFloat]
        list of the x-coordinates of the points defining the polyline.
    all_points_y : list[NonNegativeFloat]
        list of the y-coordinates of the points defining the polyline.
    name : str
        Name of shape (must be "polyline" for PolylineShapeAttributes instances).
    """

    all_points_x: list[NonNegativeFloat]
    all_points_y: list[NonNegativeFloat]
    name: str = Field("polyline", regex=r"^polyline$")

    @validator("all_points_y")
    def same_number_of_points_in_both_dimensions(cls, v, values):
        if "all_points_x" in values and len(v) != len(values["all_points_x"]):
            raise ValueError("must have same number of y points as x points")
        return v

    def as_circle(self) -> CircleShapeAttributes:
        """Calculates the smallest enclosing circle of the polyline.

        Returns
        -------
        smallest_enclosing_circle : CircleShapeAttributes
            Smallest enclosing circle of polyline.

        Examples
        --------
        >>> polyline = PolylineShapeAttributes(all_points_x=[0, 1], all_points_y=[0, 0])
        >>> polyline.as_circle()
        CircleShapeAttributes(name='circle', cx=0.5, cy=0.0, r=0.5)
        """
        cx, cy, r = smallest_enclosing_circle(zip(self.all_points_x, self.all_points_y))
        return CircleShapeAttributes(cx=cx, cy=cy, r=r)

    def get_bounding_box(self) -> BoundingBox:
        """Finds the bounding box of all the points in the polyline.

        Returns
        -------
        box : BoundingBox
            Bounding box of all the points in the polyline.

        Examples
        --------
        >>> polyline = PolylineShapeAttributes(all_points_x=[0, 1], all_points_y=[0, 0])
        >>> polyline.get_bounding_box()
        BoundingBox(x0=0, y0=0, x1=2, y1=1)
        """
        x_min = min(self.all_points_x)
        x_max = max(self.all_points_x)
        y_min = min(self.all_points_y)
        y_max = max(self.all_points_y)

        return BoundingBox(
            x0=int(x_min), y0=int(y_min), x1=int(x_max) + 1, y1=int(y_max) + 1
        )

    def in_box(self, box: BoundingBox) -> bool:
        """Returns True if all points in self are within bounding box.

        Parameters
        ----------
        box: BoundingBox
            Box to test against.

        Returns
        -------
        is_in_box : bool
            True if entire polyline is contained in box.

        Examples
        --------
        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[1, 3],
        ...     all_points_y=[15, 13],
        ... )
        >>> polyline.in_box(BoundingBox(x0=1, y0=13, x1=4, y1=16))
        True
        >>> polyline.in_box(BoundingBox(x0=2, y0=13, x1=4, y1=16))
        False
        >>> polyline.in_box(BoundingBox(x0=1, y0=14, x1=4, y1=16))
        False
        >>> polyline.in_box(BoundingBox(x0=1, y0=13, x1=3, y1=16))
        False
        >>> polyline.in_box(BoundingBox(x0=1, y0=13, x1=4, y1=15))
        False
        """
        return self.get_bounding_box().in_box(box)

    def extract_region_of_interest(
        self, image: torch.Tensor, scan_distance: PositiveInt
    ) -> torch.Tensor:
        """Extracts region of interest (ROI) from an image tensor.

        Parameters
        ----------
        image : torch.Tensor
            Image to extract region of interest from. Should be greyscale (ie. just have
            two axes)
        scan_distance : PositiveInt
            Half-width of rois for motion blurs.

        Returns
        -------
        roi : torch.Tensor
           Rotated, cropped, and straightened region of interest.

        Examples
        --------
        >>> image = torch.tensor([
        ...     [0.0, 0.1, 0.2, 0.3, 0.4],
        ...     [1.0, 1.1, 1.2, 1.3, 1.4],
        ...     [2.0, 2.1, 2.2, 2.3, 2.4],
        ...     [3.0, 3.1, 3.2, 3.3, 3.4],
        ...     [4.0, 4.1, 4.2, 4.3, 4.4],
        ... ])
        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[1, 4],
        ...     all_points_y=[2, 2],
        ... )
        >>> polyline.extract_region_of_interest(image, 1)
        tensor([[2.1000, 2.2000, 2.3000]])
        >>> polyline.extract_region_of_interest(image, 2)
        tensor([[1.1000, 1.2000, 1.3000],
                [2.1000, 2.2000, 2.3000],
                [3.1000, 3.2000, 3.3000]])

        Also works for multi-segment polylines

        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[1, 2, 4],
        ...     all_points_y=[2, 2, 2],
        ... )
        >>> polyline.extract_region_of_interest(image, 1)
        tensor([[2.1000, 2.2000, 2.3000]])
        >>> polyline.extract_region_of_interest(image, 2)
        tensor([[1.1000, 1.2000, 1.3000],
                [2.1000, 2.2000, 2.3000],
                [3.1000, 3.2000, 3.3000]])

        Segments can have different angles to each other

        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[1, 3, 3],
        ...     all_points_y=[2, 2, 0],
        ... )
        >>> polyline.extract_region_of_interest(image, 1)
        tensor([[2.1000, 2.2000, 2.3000, 1.3000]])
        >>> polyline.extract_region_of_interest(image, 2)
        tensor([[1.1000, 1.2000, 2.2000, 1.2000],
                [2.1000, 2.2000, 2.3000, 1.3000],
                [3.1000, 3.2000, 2.4000, 1.4000]])

        And segments can be at arbitrary angles. This example starts towards the top-
        right corner and travels towards the bottom-left.

        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[3, 0],
        ...     all_points_y=[1, 4],
        ... )
        >>> polyline.extract_region_of_interest(image, 2)
        tensor([[2.0778, 2.7142, 3.3506, 3.9870],
                [1.3000, 1.9364, 2.5728, 3.2092],
                [0.5222, 1.1586, 1.7950, 2.4314]])

        And this one from the top-left, heading down and left

        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[1, 4],
        ...     all_points_y=[1, 4],
        ... )
        >>> polyline.extract_region_of_interest(image, 2)
        tensor([[0.4636, 1.2414, 2.0192, 2.7971],
                [1.1000, 1.8778, 2.6556, 3.4335],
                [1.7364, 2.5142, 3.2920, 4.0698]])
        """

        def pair(items):
            return zip(items[:-1], items[1:])

        # Torch transforms need colour channel, and image needs to be padded
        img = pad(image.reshape((1,) + image.shape), scan_distance)

        sections = []
        for (x0, x1), (y0, y1) in zip(pair(self.all_points_x), pair(self.all_points_y)):
            # Correct for padding
            cx = x0 + scan_distance + 0.5
            cy = y0 + scan_distance + 0.5

            # Calculate angle of section
            rotation = atan2(y1 - y0, x1 - x0)

            # Calculate section length
            section_length = int(round(sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)))

            # Determine if image needs extra padding along x-axis
            if img.shape[-1] - cx < section_length:
                img = pad(img, [0, 0, int(section_length + cx - img.shape[-1] + 1), 0])

            # Rotate and translate image
            rotated_img = rotate(
                img,
                degrees(rotation),
                interpolation=InterpolationMode.BILINEAR,
                center=[cx, cy],
            )

            # Crop rotated image to ROI
            cropped_img = rotated_img.reshape(rotated_img.shape[1:])[
                int(cy) - scan_distance + 1 : int(cy) + scan_distance,
                int(cx) : int(cx) + section_length,
            ]

            sections.append(cropped_img)

        # Join sections to form complete ROI
        joined_img = torch.hstack(sections)

        return joined_img

    def length(self):
        """Get the sum of lengths of all the polyline segments.

        Returns
        -------
        length : float
            Length of polyline in pixels.

        Examples
        --------
        >>> from pytest import approx
        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[0, 1, 2],
        ...     all_points_y=[0, 1, 1],
        ... )
        >>> polyline.length() == approx(sqrt(2) + 1)
        True
        """
        xs = torch.tensor(self.all_points_x)
        ys = torch.tensor(self.all_points_y)
        return float(((xs[1:] - xs[:-1]) ** 2 + (ys[1:] - ys[:-1]) ** 2).sqrt().sum())

    def to_shapely(self) -> LineString:
        """Casts self to a shapely.geometry.LineString instance.

        Returns
        -------
        line_string : LineString
            Shapely representation of a polyline.

        Examples
        --------
        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[0, 1, 1],
        ...     all_points_y=[1, 1, 2],
        ... )
        >>> line_string = polyline.to_shapely()
        >>> isinstance(line_string, LineString)
        True
        >>> print(line_string)
        LINESTRING (0 1, 1 1, 1 2)
        """
        return LineString(zip(self.all_points_x, self.all_points_y))

    def hausdorff_distance(self, polyline: PolylineShapeAttributes) -> NonNegativeFloat:
        """Returns the Hausdorff distance between two PolylineShapeAttributes instances.

        Parameters
        ----------
        polyline : PolylineShapeAttributes
            Other polyline to compare to

        Returns
        -------
        h_dist : NonNegativeFloat
            Hausdorff distance between self and polyline.

        Examples
        --------
        >>> polyline0 = PolylineShapeAttributes(
        ...     all_points_x=[0, 1],
        ...     all_points_y=[0, 0],
        ... )
        >>> polyline1 = PolylineShapeAttributes(
        ...     all_points_x=[0, 1],
        ...     all_points_y=[1, 1],
        ... )
        >>> polyline0.hausdorff_distance(polyline1)
        1.0
        """
        return self.to_shapely().hausdorff_distance(polyline.to_shapely())



"""Provides utilities for working with weather data, and for calculating the movement
of the sun from the persepcitve of specific locations.

Constants defined in this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

These can all be overwritten by setting environment variables with the same name.

* ``SKYFIELD_DATA_DIR``. Path to directory to save skyfield ephemeris data. By default,
  ``~/skyfield-data`` will be used.
* ``CAMFI_EPHEMERIS``. Name of ephemeris file to use for calculating sunset and
  twilight. By default, ``de440s.bsp`` will be used. See the `choosing an ephemeris`_
  in the Skyfield documentation for possible other ephemeris files to use. Note that the
  ephemeris file will be loaded when this module is imported. The first time this
  happens, the ephemeris file will be downloaded (def4402.bsp is about 37 mb).

**Note:** For testing purposes,
a tiny abbreviated ephemeris file is included
in the Camfi git repository.
Without this, CI testing wouldn't work.
The included ephemeris limits
the date range to a few days in 2021,
so it should not be used when running Camfi normally.
To use the included ephemeris (for testing purposes only),
set the following environment variables
(assuming you are running the tests from the camfi repo root directory)::

    SKYFIELD_DATA_DIR="camfi/test/data"
    CAMFI_EPHEMERIS="test_ephem.bsp"

.. _choosing an ephemeris: https://rhodesmill.org/skyfield/planets.html#choosing-an-ephemeris
"""

from datetime import date, datetime, time, timedelta, timezone
import os
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
import pandas as pd
from pydantic import (
    BaseModel,
    FilePath,
    NonNegativeFloat,
    ValidationError,
    validator,
)
from skyfield.api import Loader, wgs84
from skyfield import almanac

from camfi.util import Timezone, Field


# Initialise skyfield
SKYFIELD_DATA_DIR = os.getenv(
    "SKYFIELD_DATA_DIR", str(Path("~/skyfield-data").expanduser())
)
CAMFI_EPHEMERIS = os.getenv("CAMFI_EPHEMERIS", "de440s.bsp")

_load = Loader(SKYFIELD_DATA_DIR)
ephemeris = _load(CAMFI_EPHEMERIS)
timescale = _load.timescale()

TWILIGHT_TRANSITIONS = {
    1: "astronomical_twilight_start",
    2: "nautical_twilight_start",
    3: "civil_twilight_start",
    4: "sunrise",
    5: "astronomical_twilight_end",
    6: "nautical_twilight_end",
    7: "civil_twilight_end",
    8: "sunset",
}


class Location(BaseModel):
    """Provides methods for working with locations.

    Parameters
    ----------
    name : str
        Name of location.
    lat : float
        Decimal latitude.
    lon : float
        Decimal longitude.
    elevation_m : NonNegativeFloat
        Elevation in metres.
    tz : timezone
        Timezone offset. Can be given as ISO8601 timezone offset str (e.g. 'Z' or
        '+10:00' or simply '+10').

    Examples
    --------
    >>> Location(
    ...     name="canberra",
    ...     lat=-35.293056,
    ...     lon=149.126944,
    ...     elevation_m=578,
    ...     tz="+10:00",
    ... )
    Location(name='canberra', lat=-35.293056, lon=149.126944, elevation_m=578.0, tz=Timezone(datetime.timezone(datetime.timedelta(seconds=36000))))
    >>> Location(
    ...     name="greenwich",
    ...     lat=51.48,
    ...     lon=0,
    ...     elevation_m=47,
    ...     tz="Z",
    ... )
    Location(name='greenwich', lat=51.48, lon=0.0, elevation_m=47.0, tz=Timezone(datetime.timezone.utc))
    >>> Location(
    ...     name="nyc",
    ...     lat=40.712778,
    ...     lon=-74.006111,
    ...     elevation_m=10,
    ...     tz="-05",
    ... )
    Location(name='nyc', lat=40.712778, lon=-74.006111, elevation_m=10.0, tz=Timezone(datetime.timezone(datetime.timedelta(days=-1, seconds=68400))))
    """

    name: str = Field(
        ..., description="Name of location. Used to link to camera placements."
    )
    lat: float = Field(..., description="Decimal latitude.")
    lon: float = Field(..., description="Decimal longitude.")
    elevation_m: NonNegativeFloat = Field(..., description="Elevation in metres.")
    tz: Timezone = Field(..., description="ISO8601 timezone offset.")

    class Config:
        schema_extra = {
            "description": "Contains spatial data on locations, including timezone."
        }

    @property
    def _dark_twilight_day(self):
        return almanac.dark_twilight_day(
            ephemeris, wgs84.latlon(self.lat, self.lon, elevation_m=self.elevation_m)
        )

    def _get_tz_aware_dt(self, dt: datetime) -> datetime:
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=self.tz._timezone)
        return dt

    def twilight_state(self, dt: datetime) -> int:
        """Gets the twilight state for the location at the specified time(s).

        The meanings of the returned integer values are

        0. Dark of night.
        1. Astronomical twilight.
        2. Nautical twilight.
        3. Civil twilight.
        4. Daytime.

        Parameters
        ----------
        dt : datetime
            datetime to evaluate. If timezone-naive, timezone will be taken from
            self.tz.

        Returns
        -------
        ts : int
            Twilight value.

        Examples
        --------
        >>> location = Location(
        ...     name="canberra",
        ...     lat=-35.293056,
        ...     lon=149.126944,
        ...     elevation_m=578,
        ...     tz="+10:00",
        ... )
        >>> location.twilight_state(datetime.fromisoformat("2021-07-28T12:00:00+10:00"))
        4
        >>> location.twilight_state(datetime.fromisoformat("2021-07-28T23:00:00+11:00"))
        0

        Timezone will be taken from self.tz if dt is timezone-naive.

        >>> location.twilight_state(datetime.fromisoformat("2021-07-28T12:00:00"))
        4

        Skyfield provides mapping from these numbers to strings.

        >>> almanac.TWILIGHTS[0]
        'Night'
        >>> almanac.TWILIGHTS[1]
        'Astronomical twilight'
        >>> almanac.TWILIGHTS[2]
        'Nautical twilight'
        >>> almanac.TWILIGHTS[3]
        'Civil twilight'
        >>> almanac.TWILIGHTS[4]
        'Day'
        """
        time = timescale.from_datetime(self._get_tz_aware_dt(dt))
        return int(self._dark_twilight_day(time))

    def twilight_states(self, datetimes: Sequence[datetime]) -> np.ndarray:
        """Like Location.twilight_state but operates on sequence of datetimes.

        Parameters
        ----------
        datetimes : Sequence[datetime]
            datetimes to evaluate. If timezone-naive, timezone will be taken from
            self.tz.

        Returns
        -------
        ts : np.ndarray
            Twilight values.

        Examples
        --------
        >>> location = Location(
        ...     name="canberra",
        ...     lat=-35.293056,
        ...     lon=149.126944,
        ...     elevation_m=578,
        ...     tz=timezone(timedelta(hours=10)),
        ... )
        >>> datetimes = [
        ...     datetime.fromisoformat("2021-07-28T12:00:00+10:00"),
        ...     datetime.fromisoformat("2021-07-28T23:00:00+11:00"),
        ...     datetime.fromisoformat("2021-07-28T12:00:00"),
        ... ]
        >>> location.twilight_states(datetimes)
        array([4, 0, 4])
        """
        datetimes = [self._get_tz_aware_dt(dt) for dt in datetimes]
        times = timescale.from_datetimes(datetimes)
        return self._dark_twilight_day(times)

    def search_sun_times(self, day: date) -> dict[str, datetime]:
        """Gets sunrise, sunset, and twilight times for a given date.

        Parameters
        ----------
        day : date
            Day to get times from.

        Returns
        -------
        twilight_times : dict[str, datetime]
            dictionary with keys "astronomical_twilight_start",
            "nautical_twilight_start", "civil_twilight_start", "sunrise", "sunset",
            "nautical_twilight_end", "civil_twilight_end", "astronomical_twilight_end".

        Examples
        --------
        >>> location = Location(
        ...     name="canberra",
        ...     lat=-35.293056,
        ...     lon=149.126944,
        ...     elevation_m=578,
        ...     tz=timezone(timedelta(hours=10)),
        ... )
        >>> day = date(2021, 7, 28)
        >>> tt = location.search_sun_times(day)

        The ordering of the transitions is as expected.

        >>> tt["astronomical_twilight_start"] < tt["nautical_twilight_start"]
        True
        >>> tt["nautical_twilight_start"] < tt["civil_twilight_start"]
        True
        >>> tt["civil_twilight_start"] < tt["sunrise"]
        True
        >>> tt["sunrise"] < tt["sunset"]
        True
        >>> tt["sunset"] < tt["civil_twilight_end"]
        True
        >>> tt["civil_twilight_end"] < tt["nautical_twilight_end"]
        True
        >>> tt["nautical_twilight_end"] < tt["astronomical_twilight_end"]
        True

        And all of the datetimes are on the correct day.

        >>> all(d.date() == day for d in tt.values())
        True
        """
        start_time = datetime.combine(date=day, time=time(0), tzinfo=self.tz._timezone)
        end_time = start_time + timedelta(days=1)
        t0 = timescale.from_datetime(start_time)
        t1 = timescale.from_datetime(end_time)

        times, twilight_types = almanac.find_discrete(t0, t1, self._dark_twilight_day)
        twilight_transitions = (
            np.roll(twilight_types, 1) > twilight_types
        ) * 5 + twilight_types

        twilight_times: dict[str, datetime] = {}
        for t, tt in zip(times, twilight_transitions):
            twilight_times[TWILIGHT_TRANSITIONS[tt]] = t.utc_datetime().astimezone(
                self.tz._timezone
            )

        return twilight_times

    def get_sun_time_dataframe(self, days: Sequence[date]) -> pd.DataFrame:
        """Calls self.search_sun_times on each day in days, and builds a DataFrame of
        sun times.

        Parameters
        ----------
        days : Sequence[date]
            Dates which will become index for dataframe.

        Returns
        -------
        sun_df : pd.DataFrame
            DataFrame indexed by location and date, with columns
            "astronomical_twilight_start", "nautical_twilight_start",
            "civil_twilight_start", "sunrise", "sunset", "nautical_twilight_end",
            "civil_twilight_end", "astronomical_twilight_end".

        Examples
        --------
        >>> location = Location(
        ...     name="canberra",
        ...     lat=-35.293056,
        ...     lon=149.126944,
        ...     elevation_m=578,
        ...     tz=timezone(timedelta(hours=10)),
        ... )
        >>> days = [date(2021, 7, 23), date(2021, 7, 24), date(2021, 7, 25)]
        >>> sun_df = location.get_sun_time_dataframe(days)
        >>> np.all(sun_df["sunset"] > sun_df["sunrise"])
        True
        >>> sun_df
                                 astronomical_twilight_start  ...        astronomical_twilight_end
        location date                                         ...
        canberra 2021-07-23 2021-07-23 05:36:52.178788+10:00  ... 2021-07-23 18:43:25.223475+10:00
                 2021-07-24 2021-07-24 05:36:20.903963+10:00  ... 2021-07-24 18:43:59.629041+10:00
                 2021-07-25 2021-07-25 05:35:48.170485+10:00  ... 2021-07-25 18:44:34.315154+10:00
        <BLANKLINE>
        [3 rows x 8 columns]
        """
        sun_times: dict[str, list[pd.Timestamp]] = {
            "astronomical_twilight_start": [],
            "nautical_twilight_start": [],
            "civil_twilight_start": [],
            "sunrise": [],
            "sunset": [],
            "nautical_twilight_end": [],
            "civil_twilight_end": [],
            "astronomical_twilight_end": [],
        }
        for day in days:
            sun_time_dict = self.search_sun_times(day)
            for key in sun_times.keys():
                sun_times[key].append(pd.Timestamp(sun_time_dict[key]))

        sun_times["date"] = [pd.Timestamp(day) for day in days]
        sun_times["location"] = [self.name for _ in range(len(days))]
        sun_df = pd.DataFrame(data=sun_times)
        sun_df.set_index(["location", "date"], inplace=True)
        return sun_df


class WeatherStation(BaseModel):
    """Contains information on a weather station.

    Parameters
    ----------
    location : Location
        Location of weather station.
    data_file : FilePath
        Path to csv file containing weather data from weather station.
    """

    location: Location = Field(..., description="Location of weather station.")
    data_file: FilePath = Field(
        ...,
        description=(
            "Path to csv file containing weather data from weather station. "
            "The firt 6 lines of the file are skipped, "
            "and the 7th should contain column headers. "
            "Should have one line per date. "
            "Minimally, the first column should be date, in YYYY-mm-dd format. "
        ),
    )

    class Config:
        schema_extra = {"description": "Contains information on a weather station."}

    def load_dataframe(self):
        """Loads weather data from self.data_file into a pd.DataFrame

        Returns
        -------
        weather_df : pd.DataFrame
            DataFrame with daily weather data, indexed by "weather_station" and "date".
        """
        weather_df = pd.read_csv(
            self.data_file, skiprows=5, header=0, parse_dates=["date"]
        )
        weather_df["weather_station"] = self.location.name
        weather_df.set_index(["weather_station", "date"], inplace=True)
        return weather_df


class LocationWeatherStationCollector(BaseModel):
    """Contains lists of Locations and Weather stations, and a mapping between them.

    Parameters
    ----------
    locations : list[Location]
        list of locations where cameras have been placed.
    weather_stations : list[WeatherStation]
        list of weather stations.
    location_weather_station_mapping : dict[str, str]
        A mapping between location names and weather_station names.
    """

    locations: list[Location] = Field(
        ..., description="list of locations where cameras have been placed."
    )
    weather_stations: list[WeatherStation] = Field(
        ..., description="list of weather stations."
    )
    location_weather_station_mapping: dict[str, str] = Field(
        ..., description="A mapping between location names and weather_station names."
    )

    class Config:
        schema_extra = {"description": "Defines Locations and Weather stations."}

    @validator("location_weather_station_mapping")
    def mapping_contains_all_locations(cls, v, values):
        if "locations" not in values:
            return v
        for location in values["locations"]:
            assert (
                location.name in v
            ), f"{location.name} missing from location_weather_station_mapping."
        return v

    @validator("location_weather_station_mapping")
    def all_weather_stations_included(cls, v, values):
        if "weather_stations" not in values:
            return v
        weather_station_location_names = set(
            ws.location.name for ws in values["weather_stations"]
        )
        for val in v.values():
            assert (
                val in weather_station_location_names
            ), f"Undefined weather station {val}. Either remove from mapping or define."
        return v

    def get_sun_time_dataframe(self, days: dict[str, Sequence[date]]) -> pd.DataFrame:
        """Calls .get_sun_time_dataframe on each location in self.locations, and builds
        a DataFrame of sun times.

        Parameters
        ----------
        days : dict[str, Sequence[date]]
            Mapping from location name to sequences of dates which will become index for
            the dataframe.

        Returns
        -------
        sun_df : pd.DataFrame
            DataFrame indexed by location and date, with columns
            "astronomical_twilight_start", "nautical_twilight_start",
            "civil_twilight_start", "sunrise", "sunset", "nautical_twilight_end",
            "civil_twilight_end", "astronomical_twilight_end".
        """
        sun_dfs = []
        for location in self.locations:
            dates = days[location.name]
            sun_dfs.append(location.get_sun_time_dataframe(dates))

        return pd.concat(sun_dfs)

    def get_weather_dataframe(self) -> pd.DataFrame:
        """Calls .load_dataframe() on each WeatherStation in self.weather_stations, and
        builds a DataFrame of weather data.

        Returns
        -------
        weather_df : pd.DataFrame
            DataFrame with daily weather data, indexed by "weather_station" and "date".
        """
        weather_dfs = []
        for weather_station in self.weather_stations:
            weather_dfs.append(weather_station.load_dataframe())

        return pd.concat(weather_dfs)

    def get_weather_sun_dataframe(
        self, days: Optional[dict[str, Sequence[date]]] = None
    ) -> pd.DataFrame:
        """Calls self.get_weather_dataframe and self.get_sun_time_dataframe, and merges
        the results into a single DataFrame.

        Parameters
        ----------
        days : Optional[dict[str, Sequence[date]]]
            Mapping from location name to sequences of dates which will become index for
            the dataframe. If None (default), this will be inferred from the weather
            dataframe.

        Returns
        -------
        weather_sun_df : pd.DataFrame
            Merged weather and sun time DataFrame.
        """
        # Get weather_df
        weather_df = self.get_weather_dataframe()

        # Infer days from self.locations and weather_df, if not already set.
        if days is None:
            days = {}
            for location in self.locations:
                weather_station: str = self.location_weather_station_mapping[
                    location.name
                ]
                days[location.name] = weather_df.loc[weather_station].index.date

        # Get sun_df
        sun_df = self.get_sun_time_dataframe(days)

        # Add "weather_station" column to sun_df to prepare for merge
        # Also add "lat", "lon", and "elevation_m" colmuns
        for location in self.locations:
            weather_station = self.location_weather_station_mapping[location.name]
            sun_df.loc[location.name, "weather_station"] = weather_station
            sun_df.loc[location.name, "lat"] = location.lat
            sun_df.loc[location.name, "lon"] = location.lon
            sun_df.loc[location.name, "elevation_m"] = location.elevation_m

        # Demote index to regular columns before merge to prevent "location" from
        # disappearing mysteriously.
        sun_df.reset_index(inplace=True)

        # Merge and reindex
        weather_sun_df = pd.merge(
            sun_df, weather_df, how="left", on=["weather_station", "date"], sort=True
        )
        weather_sun_df.set_index(["location", "date"], inplace=True)

        return weather_sun_df



"""Defines data structures relating to VGG Image Annotator. Depends on
camfi.datamodel.geometry."""

from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timedelta, tzinfo
from pathlib import Path
from typing import Any, Callable, Mapping, Optional, Union

import exif
import pandas as pd
from pydantic import (
    BaseModel,
    PositiveFloat,
    PositiveInt,
    validator,
)
import torch
import torchvision.io
from tqdm import tqdm

from .geometry import (
    BoundingBox,
    PointShapeAttributes,
    CircleShapeAttributes,
    PolylineShapeAttributes,
)
from .via_region_attributes import ViaRegionAttributes
from .region_filter_config import RegionFilterConfig
from camfi.util import DatetimeCorrector, Field


class ViaFileAttributes(BaseModel):
    """Contains file-level metadata for a single photograph.

    Parameters
    ----------
    datetime_corrected : Optional[datetime]
        Time image was taken (after taking the error of the camera's clock into
        account).
    datetime_original : Optional[datetime]
        Time image was taken (according to camera).
    exposure_time : Optional[PositiveFloat]
        The exposure time of the image in seconds, as reported by the camera.
    location : Optional[str]
        The location the image was taken.
    pixel_x_dimension : Optional[PositiveInt]
        The width of the image in pixels.
    pixel_y_dimension : Optional[PositiveInt]
        The height of the image in pixels.
    """

    datetime_corrected: Optional[datetime]
    datetime_original: Optional[datetime]
    exposure_time: Optional[PositiveFloat]
    location: Optional[str]
    pixel_x_dimension: Optional[PositiveInt]
    pixel_y_dimension: Optional[PositiveInt]

    @validator("datetime_corrected", "datetime_original", pre=True)
    def valid_datetime_str(cls, v):
        try:
            return datetime.fromisoformat(v)
        except ValueError:
            return datetime.strptime(v, "%Y:%m:%d %H:%M:%S")


class ViaRegion(BaseModel):
    """Combines region metadata with a geometry to define a complete annotation of a
    single flying insect motion blur.

    Parameters
    ----------
    region_attributes : ViaRegionAttributes
        Metadata of annotation.
    shape_attributes : Union[
        PolylineShapeAttributes, CircleShapeAttributes, PointShapeAttributes
    ]
        Geometry of annotation.
    """

    region_attributes: ViaRegionAttributes
    shape_attributes: Union[
        PolylineShapeAttributes, CircleShapeAttributes, PointShapeAttributes
    ]

    @validator("shape_attributes")
    def only_polylines_get_wingbeat_data(cls, v, values):
        if "region_attributes" in values and v.name != "polyline":
            for field in [
                values["region_attributes"].best_peak,
                values["region_attributes"].blur_length,
                values["region_attributes"].snr,
                values["region_attributes"].wb_freq_up,
                values["region_attributes"].wb_freq_down,
                values["region_attributes"].et_up,
                values["region_attributes"].et_dn,
            ]:
                assert field is None, "Wingbeat data is invalid for non-polylines"
        return v

    def get_bounding_box(self) -> BoundingBox:
        """Calls .get_bounding_box on self.shape_attributes to get the bounding box
        of the annotation.

        Returns
        -------
        box : BoundingBox
            Bounding box of annotation.
        """
        return self.shape_attributes.get_bounding_box()

    def in_box(self, box: BoundingBox) -> bool:
        """Returns True if all points in region are within bounding box.

        Parameters
        ----------
        box: BoundingBox
            Box to test against.

        Examples
        --------
        >>> polyline = PolylineShapeAttributes(
        ...     all_points_x=[1, 3],
        ...     all_points_y=[15, 13],
        ... )
        >>> region = ViaRegion(
        ...     region_attributes=ViaRegionAttributes(),
        ...     shape_attributes=polyline,
        ... )
        >>> region.in_box(BoundingBox(x0=1, y0=13, x1=4, y1=16))
        True
        >>> region.in_box(BoundingBox(x0=2, y0=13, x1=4, y1=16))
        False
        >>> region.in_box(BoundingBox(x0=1, y0=14, x1=4, y1=16))
        False
        >>> region.in_box(BoundingBox(x0=1, y0=13, x1=3, y1=16))
        False
        >>> region.in_box(BoundingBox(x0=1, y0=13, x1=4, y1=15))
        False
        """
        return self.shape_attributes.in_box(box)

    def passes_filter(self, filters: RegionFilterConfig) -> bool:
        """Determine whether self passes filters.

        Parameters
        ----------
        filters : RegionFilterConfig
            Filters to check.

        Returns
        -------
        passes : bool
           True if all filters are passed.
        """
        for field in filters.__fields__.keys():
            attribute = getattr(self.region_attributes, field)
            filt = getattr(filters, field)
            if filt.exclude_none and attribute is None:
                return False
            if attribute > filt.le or attribute < filt.ge:
                return False

        return True


class ViaMetadata(BaseModel):
    """Combines file-level image metadata with a list of annotations contained within
    the image.

    Parameters
    ----------
    file_attributes : ViaFileAttributes
        File-level image metadata.
    filename : Path
        Relative path to image file.
    regions : list[ViaRegion]
        list of flying insect annotations.
    size : int = -1
        Not used. (Included for compatability with VIA).
    """

    file_attributes: ViaFileAttributes
    filename: Path
    regions: list[ViaRegion]
    size: int = -1

    def get_bounding_boxes(self) -> list[BoundingBox]:
        """Calls .get_bounding_box on each region in self.regions.

        Returns
        -------
        boxes : list[BoundingBox]
            list of bounding boxes, with one BoundingBox per item in self.regions.
        """
        return [region.get_bounding_box() for region in self.regions]

    def get_labels(self) -> list[PositiveInt]:
        """Gets a list full of 1's with same length as self.regions.

        Returns
        -------
        labels : list[int]
            [1, 1, 1, ...]
        """
        return [1 for _ in range(len(self.regions))]

    def load_exif_metadata(
        self,
        root: Path = Path(),
        location: Optional[str] = None,
        datetime_corrector: Optional[DatetimeCorrector] = None,
    ) -> None:
        """Extract EXIF metadata from an image file and put it in
        self.file_attributes.

        **Note: this will overwrite all contents in self.file_attributes.**

        EXIF tags loaded:
          - datetime_original: datetime
          - exposure_time: PositiveFloat
          - pixel_x_dimension: PositiveInt
          - pixel_y_dimension: PositiveInt

        Extra tags:
          - datetime_corrected: datetime
                if datetime_corrector is set, this is calculated by
                calling datetime_corrector(datetime_original).
          - location: str
                set if location is set

        Parameters
        ----------
        root : Path
            Root directory from which the relative path in self.filename is resolved.
            Defaults to current working directory. If a str is passed it will be coerced
            to a Path.
        location : Optional[str]
            Option to also apply a location
        datetime_corrector : Optional[DatetimeCorrector]
            If set, then will be used to calculate datetime_corrected

        Returns
        -------
        None (operates in place)

        Examples
        --------
        >>> metadata = ViaMetadata(
        ...     file_attributes=ViaFileAttributes(),
        ...     filename="camfi/test/data/DSCF0010.JPG",
        ...     regions=[],
        ... )
        >>> metadata.load_exif_metadata()
        >>> metadata.file_attributes.datetime_original
        datetime.datetime(2019, 11, 14, 20, 30, 29)
        >>> print(round(metadata.file_attributes.exposure_time, 6))
        0.111111
        >>> metadata.file_attributes.pixel_y_dimension
        3456
        >>> metadata.file_attributes.pixel_x_dimension
        4608

        Optionally specify root directory. Here we are loading the same file, but using
        a root parameter. Note that root may also be a relative path (as in this case).
        Absolute paths are also acceptable.

        >>> metadata_with_root = ViaMetadata(
        ...     file_attributes=ViaFileAttributes(),
        ...     filename="data/DSCF0010.JPG",
        ...     regions=[],
        ... )
        >>> metadata_with_root.load_exif_metadata(root="camfi/test")
        >>> metadata_with_root.file_attributes == metadata.file_attributes
        True

        If location is set, this will be reflected

        >>> metadata = ViaMetadata(
        ...     file_attributes=ViaFileAttributes(),
        ...     filename="camfi/test/data/DSCF0010.JPG",
        ...     regions=[],
        ... )
        >>> metadata.load_exif_metadata(location="cabramurra")
        >>> metadata.file_attributes.location
        'cabramurra'

        If a time correction needs to be made (for example if the camera's clock is
        known to have been incorrectly set), then we can correct the datetime by
        supplying a function to the datetime_corrector parameter.

        >>> metadata = ViaMetadata(
        ...     file_attributes=ViaFileAttributes(),
        ...     filename="camfi/test/data/DSCF0010.JPG",
        ...     regions=[],
        ... )
        >>> metadata.load_exif_metadata(
        ...     datetime_corrector=lambda dt: dt - timedelta(days=30)
        ... )
        >>> metadata.file_attributes.datetime_original
        datetime.datetime(2019, 11, 14, 20, 30, 29)
        >>> metadata.file_attributes.datetime_corrected
        datetime.datetime(2019, 10, 15, 20, 30, 29)
        """
        with open(Path(root) / self.filename, "rb") as image_file:
            image = exif.Image(image_file)

        self.file_attributes = ViaFileAttributes(
            datetime_original=image.datetime_original,
            exposure_time=image.exposure_time,
            pixel_x_dimension=image.pixel_x_dimension,
            pixel_y_dimension=image.pixel_y_dimension,
            location=location,
        )

        if (
            datetime_corrector is not None
            and self.file_attributes.datetime_original is not None
        ):
            self.file_attributes.datetime_corrected = datetime_corrector(
                self.file_attributes.datetime_original
            )

    def read_image(self, root: Path = Path()) -> torch.Tensor:
        """Read an image from a file.

        Parameters
        ----------
        root: Path
            Root directory from which the relative path in self.filename is resolved.
            Defaults to current working directory. If a str is passed it will be coerced
            to a Path.

        Returns
        -------
        torch.Tensor[colour, height (y), width (x)]
            Image as RGB float32 tensor.

        Examples
        --------
        >>> metadata = ViaMetadata(
        ...     file_attributes=ViaFileAttributes(),
        ...     filename="camfi/test/data/DSCF0010.JPG",
        ...     regions=[],
        ... )
        >>> image = metadata.read_image()
        >>> image.shape == (3, 3456, 4608)
        True
        >>> image.dtype
        torch.float32

        Optionally specify root directory. Here we are loading the same file, but using
        a root parameter. Note that root may also be a relative path (as in this case).
        Absolute paths are also acceptable.

        >>> metadata_with_root = ViaMetadata(
        ...     file_attributes=ViaFileAttributes(),
        ...     filename="data/DSCF0010.JPG",
        ...     regions=[],
        ... )
        >>> image_with_root = metadata_with_root.read_image(root="camfi/test")
        >>> image_with_root.allclose(image)
        True
        """
        image = torchvision.io.read_image(
            str(Path(root) / self.filename), mode=torchvision.io.image.ImageReadMode.RGB
        )
        return image / 255  # Converts from uint8 to float32

    def filter_regions(self, region_filters: RegionFilterConfig) -> None:
        """Filters regions in-place.

        Parameters
        ----------
        region_filters : RegionFilterConfig
            Filters to apply.
        """
        self.regions = list(
            filter(lambda x: x.passes_filter(region_filters), self.regions)
        )


class ViaProject(BaseModel):
    """Defines the structure of a VIA project file. Can be used for loading and saving
    VIA project data.

    Parameters
    ----------
    via_attributes : dict
        Unused by camfi. (Included for compatability with VIA).
    via_img_metadata : dict[str, ViaMetadata]
        dict of {str: ViaMetadata} pairs. Keys can be arbitrary strings, however they
        usually bare some resemblance to the .filename attribute of the ViaMetadata
        instance.
    via_settings : dict
        Unused by camfi. (Included for compatability with VIA).
    """

    via_attributes: dict
    via_img_metadata: dict[str, ViaMetadata]
    via_settings: dict

    class Config:
        """Sets pydantic.BaseModel configuration of ViaProject."""

        fields = {
            "via_attributes": "_via_attributes",
            "via_img_metadata": "_via_img_metadata",
            "via_settings": "_via_settings",
        }

    def load_all_exif_metadata(
        self,
        root: Path = Path(),
        location_dict: Optional[Mapping[Path, Optional[str]]] = None,
        datetime_correctors: Optional[
            Mapping[Path, Optional[DatetimeCorrector]]
        ] = None,
        disable_progress_bar: Optional[bool] = True,
    ) -> None:
        """Calls the .load_exif_metadata method on all ViaMetadata instances in
        self.via_img_metadata, extracting the EXIF metadata from each image file.

        Parameters
        ----------
        root : Path
            Root directory from which the relative path in self.filename is resolved.
            Defaults to current working directory. If a str is passed it will be coerced
            to a Path.
        location_dict : Optional[Mapping[Path, Optional[str]]]
            A mapping from filenames (i.e. relative paths to images under root) to
            location strings, which are passed to ViaMetadata.load_exif_metadata.
            Typically, an instance of camfi.util.SubDirdict should be used.
        datetime_correctors : Optional[Mapping[Path, Optional[DatetimeCorrector]]]
            A mapping from filenames (i.e. relative paths to images under root) to
            DatetimeCorrector instances, which are passed to
            ViaMetadata.load_exif_metadata
            Typically, an instance of camfi.util.SubDirdict should be used.
        disable_progress_bar : Optional[bool]
            If True (default), progress bar is disabled.
            If set to None, disable on non-TTY.

        Returns
        -------
        None
            Operates in place.

        Examples
        --------
        >>> with open("camfi/test/data/sample_project_images_included.json") as f:
        ...     project = ViaProject.parse_raw(f.read())

        The file which has been loaded contains no metadata

        >>> for meta in project.via_img_metadata.values():
        ...     print(meta.filename, str(meta.file_attributes.datetime_original))
        DSCF0010.JPG None
        DSCF0011.JPG None

        After load_all_exif_metadata is called, project does contain image metadata

        >>> project.load_all_exif_metadata(root=Path("camfi/test/data"))
        >>> for meta in project.via_img_metadata.values():
        ...     print(meta.filename, str(meta.file_attributes.datetime_original))
        DSCF0010.JPG 2019-11-14 20:30:29
        DSCF0011.JPG 2019-11-14 20:40:32

        If location_dict and/or datetime_correctors are set, the metadata will
        include location and/or datetime_corrected, respectively. Normally, these
        would be set as instances of camfi.util.SubDirdict, but for brevity we use
        a regular dict for each of them here.

        >>> project.load_all_exif_metadata(
        ...     root=Path("camfi/test/data"),
        ...     location_dict={
        ...         Path("DSCF0010.JPG"): "loc0", Path("DSCF0011.JPG"): "loc1"
        ...     },
        ...     datetime_correctors={
        ...         Path("DSCF0010.JPG"): lambda dt: dt + timedelta(hours=1),
        ...         Path("DSCF0011.JPG"): lambda dt: dt - timedelta(hours=1),
        ...     },
        ... )
        >>> for meta in project.via_img_metadata.values():
        ...     print(
        ...         meta.filename,
        ...         meta.file_attributes.location,
        ...         str(meta.file_attributes.datetime_corrected),
        ...     )
        DSCF0010.JPG loc0 2019-11-14 21:30:29
        DSCF0011.JPG loc1 2019-11-14 19:40:32
        """
        if location_dict is None:
            location_dict = defaultdict(lambda: None)
        if datetime_correctors is None:
            datetime_correctors = defaultdict(lambda: None)

        for metadata in tqdm(
            self.via_img_metadata.values(),
            disable=disable_progress_bar,
            desc="Loading EXIF metadata",
            unit="img",
            dynamic_ncols=True,
            ascii=True,
            color="green",
        ):
            metadata.load_exif_metadata(
                root=root,
                location=location_dict[metadata.filename],
                datetime_corrector=datetime_correctors[metadata.filename],
            )

    def to_region_dataframe(self) -> pd.DataFrame:
        """Returns a Pandas DataFrame with one row per region (annotation).

        Returns
        -------
        regions : pd.DataFrame
            DataFrame with a row for every annotation in self.via_img_metadata. Contains
            a column for every field in ViaFileAttributes and ViaRegionAttributes, as
            well as img_key, filename, and name columns.
        """
        rows: list[dict[str, Any]] = []
        for img_key, metadata in self.via_img_metadata.items():
            for region in metadata.regions:
                row = {
                    "img_key": img_key,
                    "filename": metadata.filename,
                    "name": region.shape_attributes.name,
                }
                row.update(dict(metadata.file_attributes))
                row.update(dict(region.region_attributes))
                rows.append(row)

        return pd.DataFrame(rows)

    def to_image_dataframe(self, tz: Union[None, str, tzinfo] = None) -> pd.DataFrame:
        """Returns a Pandas DataFrame with one row per image.

        Parameters
        ----------
        tz: Optional[tzinfo]
           If set, all datetime_corrected values will be converted to the specified
           timezone.

        Returns
        -------
        df : pd.DataFrame
            DataFrame with one row per image. Contains columns for each field in
            ViaFileAttributes, as well as img_key, filename, and n_annotations.
            Note: values in datetime_corrected (but not datetime_original) will be
            converted to pandas.Timestamp, with timezone conversion/localization applied
            if applicable.
        """
        rows: list[dict[str, Any]] = []
        for img_key, metadata in self.via_img_metadata.items():
            row = {
                "img_key": img_key,
                "filename": metadata.filename,
                "n_annotations": len(metadata.regions),
            }
            row.update(dict(metadata.file_attributes))

            # Convert datetime to pandas.Timestamp before putting in the DataFrame
            row["datetime_corrected"] = pd.to_datetime(row["datetime_corrected"])
            assert isinstance(row["datetime_corrected"], pd.Timestamp)

            if row["datetime_corrected"] is not None and tz is not None:
                # Fix timezone
                if row["datetime_corrected"].tzinfo is not None:
                    row["datetime_corrected"] = row["datetime_corrected"].astimezone(tz)
                else:
                    row["datetime_corrected"] = row["datetime_corrected"].tz_localize(
                        tz
                    )

            rows.append(row)

        return pd.DataFrame(rows)

    def filtered_copy(
        self, function: Callable[[ViaMetadata], bool], deep: bool = False
    ) -> ViaProject:
        """Filters images in self.via_img_metadata, returning a new ViaProject instance.

        Parameters
        ----------
        function : Callable[[ViaMetadata], bool]
            Called on each value in self.via_img_metadata to determine if it should be
            included in output.
        deep : bool
            If True, make a deep copy.

        Returns
        -------
        project : ViaProject
            Copy of self with via_img_metadata filtered.
        """
        if deep:
            filtered_img_metadata = {
                key: value.copy(deep=True)
                for key, value in self.via_img_metadata.items()
                if function(value)
            }
        else:
            filtered_img_metadata = {
                key: value
                for key, value in self.via_img_metadata.items()
                if function(value)
            }
        return self.copy(update={"via_img_metadata": filtered_img_metadata}, deep=deep)

    def filter_inplace(self, function: Callable[[ViaMetadata], bool]) -> None:
        """Filters images in self.via_img_metadata in-place.

        Parameters
        ----------
        function : Callable[[ViaMetadata], bool]
            Called on each value in self.via_img_metadata to determine if it should be
            included in output.
        """
        _new_img_metadata = {
            key: value
            for key, value in self.via_img_metadata.items()
            if function(value)
        }
        self.via_img_metadata = _new_img_metadata



"""Camera-based Analysis and Monitoring of Flying Insects

See:
https://github.com/J-Wall/camfi
"""

import codecs
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()


def read(rel_path):
    here = pathlib.Path(__file__).parent.resolve()
    with codecs.open(here.joinpath(rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="camfi",
    version=get_version("camfi/_version.py"),
    description="Camera-based Analysis and Monitoring of Flying Insects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/J-Wall/camfi",
    author="Jesse Wallace",
    author_email="jesse.wallace@anu.edu.au",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Image Processing",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a list of additional keywords, separated
    # by commas, to be used to assist searching for the distribution in a
    # larger catalog.
    keywords="ecology, monitoring, insects, wingbeat frequency",  # Optional
    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    # package_dir={'': 'src'},  # Optional
    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    packages=find_packages(),
    # py_modules=["camfi"],
    #
    # packages=find_packages(where='src'),  # Required
    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. See
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.9, <4",
    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "bces",
        "exif",
        "numpy",
        "pandas",
        "pydantic",
        "scikit-image",
        "scipy",
        "Shapely",
        "skyfield",
        "strictyaml",
        "torch",
        "torchvision",
        "tqdm",
    ],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    entry_points={  # Optional
        "console_scripts": [
            "camfi=camfi.cli:main",
        ],
    },
)



from datetime import datetime, timedelta, timezone, tzinfo
from functools import wraps
import itertools
from math import sqrt
from pathlib import Path
from textwrap import wrap as _wrap
from typing import (
    Callable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

import numpy as np
from pydantic import Field as _pydantic_field, NonNegativeInt
import torch


DatetimeCorrector = Callable[[datetime], datetime]


def fill(text, width=50, **kwargs):
    return "\n\n".join(_wrap(text, width=width, **kwargs))


@wraps(_pydantic_field)
def Field(*args, **kwargs):
    if "description" in kwargs:
        kwargs["description"] = fill(kwargs["description"])
    return _pydantic_field(*args, **kwargs)


# Hack to get cache decorator to play nice with mypy
T = TypeVar("T")


def _sec_trivial(points: Sequence[tuple[float, float]]) -> tuple[float, float, float]:
    if len(points) == 3:
        (x1, y1), (x2, y2), (x3, y3) = points
        A = np.array([[x3 - x1, y3 - y1], [x3 - x2, y3 - y2]])
        Y = np.array(
            [
                (x3 ** 2 + y3 ** 2 - x1 ** 2 - y1 ** 2),
                (x3 ** 2 + y3 ** 2 - x2 ** 2 - y2 ** 2),
            ]
        )
        if np.linalg.det(A) == 0:
            min_point = min(points)
            max_point = max(points)
            return (
                0.5 * (min_point[0] + max_point[0]),
                0.5 * (min_point[1] + max_point[1]),
                0.5
                * sqrt(
                    (min_point[0] - max_point[0]) ** 2
                    + (min_point[1] - max_point[1]) ** 2
                ),
            )
        Ainv = np.linalg.inv(A)
        X = 0.5 * np.dot(Ainv, Y)
        return X[0], X[1], sqrt((X[0] - x1) ** 2 + (X[1] - y1) ** 2)
    elif len(points) == 2:
        return (
            0.5 * (points[0][0] + points[1][0]),
            0.5 * (points[0][1] + points[1][1]),
            0.5
            * sqrt(
                (points[0][0] - points[1][0]) ** 2 + (points[0][1] - points[1][1]) ** 2
            ),
        )
    elif len(points) == 1:
        return points[0][0], points[0][1], 0.0
    elif len(points) == 0:
        return 0.0, 0.0, 0.0
    else:
        raise ValueError(f"{len(points)} points given. Maximum for trivial case is 3.")


def smallest_enclosing_circle(
    points: Union[Iterable[tuple[float, float]], np.ndarray]
) -> tuple[float, float, float]:
    """Performs Welzl's algorithm to find the smallest enclosing circle of a set of
    points in a cartesian plane.

    Parameters
    ----------
    points : Union[Iterable[tuple[float, float]], np.ndarray]
        Iterable of 2-tuples or (N, 2)-array, with each tuple defining the coordinates
        of a point.

    Returns
    -------
    x : float
        x-coordinate of centre of circle.
    y : float
        y-coordinate of centre of circle.
    r : float
        Radius of circle.

    Examples
    --------
    If no points are given, values are still returned:

    >>> smallest_enclosing_circle([])
    (0.0, 0.0, 0.0)

    If one point is given, r will be 0.0:

    >>> smallest_enclosing_circle([(1.0, 2.0)])
    (1.0, 2.0, 0.0)

    Two points trivial case:

    >>> smallest_enclosing_circle([(0.0, 0.0), (2.0, 0.0)])
    (1.0, 0.0, 1.0)

    Three points trivial case:

    >>> np.allclose(
    ...     smallest_enclosing_circle([(0.0, 0.0), (2.0, 0.0), (1.0, sqrt(3))]),
    ...     (1.0, sqrt(3) / 3, 2 * sqrt(3) / 3)
    ... )
    True

    Extra points within the circle don't affect the circle:

    >>> np.allclose(
    ...     smallest_enclosing_circle([
    ...                                (0.0, 0.0),
    ...                                (2.0, 0.0),
    ...                                (1.0, sqrt(3)),
    ...                                (0.5, 0.5)]),
    ...     (1.0, sqrt(3) / 3, 2 * sqrt(3) / 3)
    ... )
    True

    If points are inscribed on a circle, the correct circle is also given:

    >>> np.allclose(
    ...     smallest_enclosing_circle([(0.0, 0.0), (2.0, 0.0), (2.0, 2.0), (0.0, 2.0)]),
    ...     (1.0, 1.0, sqrt(2))
    ... )
    True
    """

    def welzl(P, R):
        if len(P) == 0 or len(R) == 3:
            return _sec_trivial(R)
        p = P[0]
        D = welzl(P[1:], R)
        if (D[0] - p[0]) ** 2 + (D[1] - p[1]) ** 2 < D[2] ** 2:
            return D

        return welzl(P[1:], R + (p,))

    P = [tuple(p) for p in points]

    return welzl(tuple(P), ())


def dilate_idx(
    rr: np.ndarray,
    cc: np.ndarray,
    d: int,
    img_shape: Optional[tuple[int, int]] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Takes index arrays rr and cc and performs a morphological dilation of size d on
    them.

    Parameters
    ----------
    rr : np.ndarray
        Row indices.
    cc : np.ndarray
        Column indices (must have same shape as rr).
    d : int
        Dilation factor, must be at least 1 (or a ValueError is raised).
    img_shape : Optional[tuple[int, int]]
        Shape of image (rows, columns). Indices which lie outside this will be ommitted.

    Returns
    -------
    rr_dilated : np.ndarray
        Row indices after morphological dilation.
    cc_dilated : np.ndarray
        Column indices after morphological dilation.

    Examples
    --------
    >>> a = np.array([[0, 0, 0, 0, 0],
    ...               [0, 0, 1, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0]])
    >>> rr, cc = np.nonzero(a)
    >>> rr_dilated, cc_dilated = dilate_idx(rr, cc, 1)
    >>> a[rr_dilated, cc_dilated] = 1
    >>> a
    array([[0, 0, 1, 0, 0],
           [0, 1, 1, 1, 0],
           [0, 0, 1, 0, 0],
           [0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0]])

    If shape is given, omits indices larger than the dimensions given

    >>> a = np.array([[0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 1],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0]])
    >>> rr, cc = np.nonzero(a)
    >>> rr_dilated, cc_dilated = dilate_idx(rr, cc, 1, a.shape)
    >>> a[rr_dilated, cc_dilated] = 1
    >>> a
    array([[0, 0, 0, 0, 1],
           [0, 0, 0, 1, 1],
           [0, 0, 0, 0, 1],
           [0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0]])

    If we didn't give the shape argument in the above example, we get an IndexError

    >>> a = np.array([[0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 1],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0]])
    >>> rr, cc = np.nonzero(a)
    >>> rr_dilated, cc_dilated = dilate_idx(rr, cc, 1)
    >>> a[rr_dilated, cc_dilated] = 1
    Traceback (most recent call last):
    ...
    IndexError: index 5 is out of bounds for axis 1 with size 5

    But we don't need the shape parameter to filter out negative indices

    >>> a = np.array([[1, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0]])
    >>> rr, cc = np.nonzero(a)
    >>> rr_dilated, cc_dilated = dilate_idx(rr, cc, 1)
    >>> a[rr_dilated, cc_dilated] = 1
    >>> a
    array([[1, 1, 0, 0, 0],
           [1, 0, 0, 0, 0],
           [0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0]])

    Dilation is based on euclidean distance

    >>> a = np.array([[0, 0, 0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0, 0, 0],
    ...               [0, 0, 0, 1, 0, 0, 0],
    ...               [0, 0, 0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0, 0, 0],
    ...               [0, 0, 0, 0, 0, 0, 0]])
    >>> rr, cc = np.nonzero(a)
    >>> rr_dilated, cc_dilated = dilate_idx(rr, cc, 3, a.shape)
    >>> a[rr_dilated, cc_dilated] = 1
    >>> a
    array([[0, 0, 0, 1, 0, 0, 0],
           [0, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 0],
           [1, 1, 1, 1, 1, 1, 1],
           [0, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 0],
           [0, 0, 0, 1, 0, 0, 0]])

    If input is sorted, then the ouptut will be too (with precedence rr, cc)

    >>> rr, cc = np.array([50]), np.array([50])
    >>> dilate_idx(rr, cc, 1, (100, 100))
    (array([49, 50, 50, 50, 51]), array([50, 49, 50, 51, 50]))

    If a non-positive dilation factor is given, a ValueError is raised

    >>> dilate_idx(1, 2, 0)
    Traceback (most recent call last):
    ...
    ValueError: d=0. Should be positive.
    """
    if d < 1:
        raise ValueError(f"d={d}. Should be positive.")

    d2 = d * d
    offset_r, offset_c = zip(
        *itertools.filterfalse(
            lambda x: x[0] ** 2 + x[1] ** 2 > d2,
            itertools.product(range(-d, d + 1), repeat=2),
        )
    )
    rr_dilated = np.stack([rr + i for i in offset_r]).ravel()
    cc_dilated = np.stack([cc + i for i in offset_c]).ravel()
    mask = rr_dilated >= 0
    mask[cc_dilated < 0] = False
    if img_shape is not None:
        mask[rr_dilated >= img_shape[0]] = False
        mask[cc_dilated >= img_shape[1]] = False

    return rr_dilated[mask], cc_dilated[mask]


V = TypeVar("V")


class SubDirDict(Mapping[Path, V]):
    """A mapping from subdirectory Paths to V which returns self['foo/bar'] if
    'foo/bar/baz' is missing from the available keys.

    Examples
    --------
    >>> d = SubDirDict()
    >>> d["foo"] = "foo"
    >>> d["foo"]
    'foo'
    >>> d["foo/bar"]
    'foo'
    >>> d["foo/bar/baz"]
    'foo'
    >>> d["bar"]
    Traceback (most recent call last):
    ...
    KeyError: "'bar' not in SubDirDict({Path('foo'): 'foo'})"

    SubDirDict can be initialised from a dictionary

    >>> SubDirDict({"foo": "bar", "foobar": "baz"})
    SubDirDict({Path('foo'): 'bar', Path('foobar'): 'baz'})
    """

    def __init__(self, mapping: Optional[Mapping[Path, V]] = None):
        """Initialises SubDirDict.

        Parameters
        ----------
        mapping: Optional[Mapping[Path, V]]
            E.g. an instance of type dictt[Path, V]
        """
        self._lastkey = None
        self._prevkey = None
        self._dict: dict[Path, V] = {}
        if mapping is not None:
            self._dict.update(mapping)

    def __getitem__(self, key):
        if isinstance(key, str):
            key = Path(key)
        elif not isinstance(key, Path):
            raise TypeError(
                f"SubDirDict can only be indexed by Path instances. Got {type(key)}"
            )
        self._prevkey = self._lastkey
        self._lastkey = key
        try:
            return self._dict[key]
        except KeyError:
            return self.__missing__(key)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            key = Path(key)
        elif not isinstance(key, Path):
            raise TypeError(
                f"SubDirDict can only be indexed by Path instances. Got {type(key)}"
            )
        self._dict[key] = value

    def __missing__(self, key):
        if key == Path():
            raise KeyError(f"'{self._prevkey}' not in {str(self)}")

        return self[key.parent]

    def __repr__(self):
        """String representation of SubDirDict. Uses "Path" instead of platform
        dependant "PosixPath" or "WindowsPath".
        """
        s = ", ".join(
            (f"Path('{str(path)}'): {value!r}" for path, value in self._dict.items())
        )
        return f"SubDirDict({{{s}}})"

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()


def endpoint_truncate(
    fit_mask_vals: np.ndarray, n: NonNegativeInt
) -> tuple[NonNegativeInt, NonNegativeInt]:
    """An implementation of an endpoint_method.

    Parameters
    ----------
    fit_mask_vals : np.ndarray
        Array to find the endpoints of.
    n : NonNegativeInt
        Number of values to truncate off start and end of fit_mask_vals.

    Returns
    -------
    start_index : NonNegativeInt
        Index of endpoint of polyline annotation.
    end_index : NonNegativeInt
        Index of endpoint of polyline annotation.
    """
    return (n, len(fit_mask_vals) - n)


endpoint_methods = {"truncate": endpoint_truncate}


def weighted_intersection_over_minimum(
    mask0: torch.Tensor, mask1: torch.Tensor
) -> float:
    """Calculates the weighted intersection over minimum (IoM) between two segmentation
    masks.

    Parameters
    ----------
    mask0 : torch.Tensor
        Instance segmentation mask to compare to mask1.
    mask1 : torch.Tensor
        Instance segmentation mask to compare to mask0. Should have the same shape as
        mask0.

    Returns
    -------
    iom : float
        Weighted intersection over union of mask0 and mask1.

    Examples
    --------
    >>> mask0 = torch.tensor([0.0, 0.5, 0.5, 0.0])
    >>> mask1 = torch.tensor([1.0, 1.0, 0.0, 0.0])
    >>> weighted_intersection_over_minimum(mask0, mask1)
    0.5
    """
    return float(torch.minimum(mask0, mask1).sum()) / min(
        float(mask0.sum()), float(mask1.sum())
    )


def parse_timezone(value: str) -> timezone:
    if value == "Z":
        return timezone.utc

    offset_mins = int(value[-2:]) if len(value) > 3 else 0
    offset = 60 * int(value[1:3]) + offset_mins
    if value[0] == "-":
        offset = -offset
    return timezone(timedelta(minutes=offset))


def encode_timezone(tz: Optional[tzinfo]) -> Optional[str]:
    if tz is None:
        return None
    if tz == timezone.utc:
        return "Z"
    tot_seconds = int(tz.utcoffset(None).total_seconds())  # type: ignore[union-attr]
    hours, minutes = divmod(abs(tot_seconds) // 60, 60)
    sign = "-" if tot_seconds < 0 else "+"
    return f"{sign}{hours:02}:{minutes:02}"


class Timezone(tzinfo):
    """Provides pydantic validation for timezones."""

    _timezone: timezone

    def __init__(self, v):
        if isinstance(v, str):
            self._timezone = parse_timezone(v)
        elif isinstance(v, timezone):
            self._timezone = v
        else:
            self._timezone = timezone(v)

    def utcoffset(self, dt: Optional[datetime]) -> Optional[timedelta]:
        return self._timezone.utcoffset(dt)

    def dst(self, dt: Optional[datetime]) -> Optional[timedelta]:
        return self._timezone.dst(dt)

    def tzname(self, dt: Optional[datetime]) -> Optional[str]:
        return self._timezone.tzname(dt)

    def fromutc(self, dt: datetime) -> datetime:
        if dt.tzinfo is not self:
            raise ValueError
        return self._timezone.fromutc(dt.replace(tzinfo=self._timezone)).replace(
            tzinfo=self
        )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            title="Timezone",
            pattern=r"^Z|[+-]\d{2}(?::?\d{2})?$",
            examples=["Z", "+10:00", "-05"],
            type="string",
        )

    @classmethod
    def validate(cls, v):
        return cls(v)

    def __str__(self):
        return encode_timezone(self._timezone)

    def __repr__(self):
        return f"Timezone({self._timezone!r})"



"""Provides classes for operating on location and time information for camera
placements. Depends on camfi.util."""

from __future__ import annotations

from datetime import datetime
from math import fsum
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel, root_validator, validator

from camfi.util import DatetimeCorrector, SubDirDict, Field


class LocationTime(BaseModel):
    """Defines camera placement data. Specifically, it includes camera placement time
    and camera retrieval time (both according to the camera's clock, and real time), as
    well as a string specifying the location the camera was placed. Provides a the
    .corrector method, which returns a DatetimeCorrector function for inferring
    real time from camera timestamps (useful if cameras have inaccurate clocks).

    Parameters
    ----------
    camera_start_time : datetime
        Timestamp from when the camera was placed (according to the camera's clock).
    actual_start_time : Optional[datetime]
        Actual timestamp from when the camera was placed (real time).
        Defaults to camera_start_time.
    camera_end_time : Optional[datetime]
        Timestamp from when the camera was retrieved (according to the camera's clock).
    actual_end_time : Optional[datetime]
        Actual timestamp from when the camera was retrieved (real time).
    location : Optional[str]
        String specifying the location the camera was placed in.
    """

    camera_start_time: datetime = Field(
        ..., description="Camera placement datetime (according to camera's clock)."
    )
    actual_start_time: Optional[datetime] = Field(
        None,
        description="Actual camera placement datetime (defaults to camera_start_time).",
    )
    camera_end_time: Optional[datetime] = Field(
        None, description="Camera retrieval time (according to camera's clock)."
    )
    actual_end_time: Optional[datetime] = Field(
        None, description="Actual camera retrieval time."
    )
    location: Optional[str] = Field(
        None, description="Name of location the camera was placed in."
    )

    class Config:
        json_encoders = {
            datetime: lambda v: v.isoformat(),
        }

        @staticmethod
        def schema_extra(schema: dict[str, Any], model: LocationTime) -> None:
            # Remove title from schema
            schema.pop("title", None)
            schema["description"] = "Camera placement data"

    @validator("actual_start_time", always=True)
    def default_actual_start_time(cls, v, values):
        if v is None and "camera_start_time" in values:
            return values["camera_start_time"]
        return v

    @root_validator
    def all_offset_aware_or_naive(cls, values):
        is_offset_aware = values["camera_start_time"].tzinfo is not None
        for field in ["actual_start_time", "camera_end_time", "actual_end_time"]:
            if field in values and values[field] is not None:
                assert (
                    values[field].tzinfo is not None
                ) == is_offset_aware, (
                    "Unable to mix timezone offset-aware and -naive datetimes"
                )
        return values

    def get_time_ratio(self) -> Optional[float]:
        """If self.camera_end_time and self.actual_end_time are set, gets the camera
        time to actual time ratio. Otherwise, None

        Returns
        -------
        camera_time_to_actual_time_ratio : float
            (Time elapsed on cameras cloack) / (Real time elapsed)

        Examples
        --------
        >>> location_time = LocationTime(
        ...     camera_start_time="2021-07-15T14:00",
        ...     camera_end_time="2021-07-15T16:00",
        ...     actual_end_time="2021-07-15T15:00",
        ... )
        >>> location_time.get_time_ratio()
        2.0

        camera_end_time and actual_end_time must be set, or else None is returned.

        >>> print(LocationTime(camera_start_time="2021-07-15T14:00").get_time_ratio())
        None
        """
        if (
            self.actual_start_time is None
            or self.camera_end_time is None
            or self.actual_end_time is None
        ):
            return None

        camera_elapsed_time = self.camera_end_time - self.camera_start_time
        actual_elapsed_time = self.actual_end_time - self.actual_start_time
        return camera_elapsed_time / actual_elapsed_time

    def corrector(
        self, camera_time_to_actual_time_ratio: Optional[float] = None
    ) -> DatetimeCorrector:
        """Returns a datetime corrector function, which takes an original camera-
        reported datetime as an argument, and returns a corrected datetime.
        If self.actual_start_time is None, then it is assumed that
        self.camera_start_time reflects the actual time. Generally it is advised to
        set the time of the camera just before it is placed out, which is the basis for
        this assumption.

        Parameters
        ----------
        camera_time_to_actual_time_ratio : Optional[float]
            The amount of time elapsed as reported by the camera divided by the actual
            amount of time elapsed. If None, then this is inferred from the fields of
            the LocationTime instance. In this case, both self.camera_end_time and
            self.actual_end_time must be set.

        Returns
        -------
        datetime_corrector : DatetimeCorrector
            Function which maps camera time to real time.

        Examples
        --------
        >>> location_time = LocationTime(camera_start_time="2021-07-15T14:00")
        >>> location_corrector = location_time.corrector(2.)
        >>> location_corrector(datetime(2021, 7, 15, 16, 0))
        datetime.datetime(2021, 7, 15, 15, 0)

        Also works with offset-aware datetimes.

        >>> location_time = LocationTime(camera_start_time="2021-07-15T14:00+10")
        >>> location_corrector = location_time.corrector(2.)
        >>> location_corrector(datetime.fromisoformat("2021-07-15T17:00+11:00"))
        datetime.datetime(2021, 7, 15, 15, 0, tzinfo=datetime.timezone(datetime.timedelta(seconds=36000)))

        If self.camera_start_time is offset aware, but the argument to the
        DatetimeCorrector function is not, it will be assumed that the argument has the
        same offset as self.camera_start_time.

        >>> location_time = LocationTime(camera_start_time="2021-07-15T14:00+10")
        >>> location_corrector = location_time.corrector(2.)
        >>> location_corrector(datetime.fromisoformat("2021-07-15T16:00"))
        datetime.datetime(2021, 7, 15, 15, 0, tzinfo=datetime.timezone(datetime.timedelta(seconds=36000)))

        However, the converse raises a TypeError.

        >>> location_time = LocationTime(camera_start_time="2021-07-15T14:00")
        >>> location_corrector = location_time.corrector(2.)
        >>> location_corrector(datetime.fromisoformat("2021-07-15T16:00+10:00"))
        Traceback (most recent call last):
        ...
        TypeError: Cannot call offset-naive DatetimeCorrector with offset-aware datetime.

        Raises an error if camera_time_to_actual_time_ratio cannot be determined.

        >>> location_time = LocationTime(camera_start_time="2021-07-15T14:00")
        >>> location_corrector = location_time.corrector()
        Traceback (most recent call last):
        ...
        ValueError: Must set camera_time_to_actual_time_ratio or both end times
        """
        if camera_time_to_actual_time_ratio is None:
            camera_time_to_actual_time_ratio = self.get_time_ratio()

        if camera_time_to_actual_time_ratio is None:
            raise ValueError(
                "Must set camera_time_to_actual_time_ratio or both end times"
            )

        def datetime_corrector(datetime_original: datetime) -> datetime:
            # To quash some mypy errors but still allow proper type checking:
            assert isinstance(camera_time_to_actual_time_ratio, float)
            assert isinstance(self.actual_start_time, datetime)

            if datetime_original.tzinfo is None:
                datetime_original = datetime_original.replace(
                    tzinfo=self.camera_start_time.tzinfo
                )
            elif self.camera_start_time.tzinfo is None:
                raise TypeError(
                    "Cannot call offset-naive DatetimeCorrector with offset-aware datetime."
                )

            camera_elapsed_time = datetime_original - self.camera_start_time
            actual_elapsed_time = camera_elapsed_time / camera_time_to_actual_time_ratio
            return self.actual_start_time + actual_elapsed_time

        return datetime_corrector


class LocationTimeCollector(BaseModel):
    """Used to generate SubDirDict instances which map subdirectories to location
    strings or DatetimeCorrectors.

    Parameters
    -----------
    camera_placements: dict[str, LocationTime]
        dictionary mapping directories to LocationTime instances. For example, you
        might have one LocationTime for each camera placement, and each camera placement
        also has it's own directory for images.

    Examples
    --------
    Here, LocationTimeCollector is instantiated with two LocationTime instances.
    The first has enough information to get a float from its .get_time_rato method,
    However the second doesn't.

    >>> lt_collector = LocationTimeCollector(camera_placements={
    ...     "data": LocationTime(
    ...         camera_start_time="2021-07-15T14:00",
    ...         camera_end_time="2021-07-15T16:00",
    ...         actual_end_time="2021-07-15T15:00",
    ...     ),
    ...     "foo": LocationTime(camera_start_time="2021-07-15T13:00"),
    ... })

    When calling the .get_time_ratio method on a LocationTimeCollector instance,
    the mean of all (excluding those who return None) .get_time_ratio results from
    all the LocationTimes is given.

    >>> lt_collector.get_time_ratio()
    2.0
    """

    camera_placements: dict[str, LocationTime] = Field(
        ...,
        description="Mapping from sub-directories containing images to LocationTime instances.",
    )

    class Config:
        schema_extra = {"description": "Contains data for multiple camera placements."}

    def get_time_ratio(self) -> Optional[float]:
        """Gets the mean of calling the .get_time_ratio method on each LocationTime
        in self.camera_placements.

        Returns
        -------
        camera_time_to_actual_time_ratio : float
            The mean of (Time elapsed on cameras cloack) / (Real time elapsed) for all
            camera placements (LocationTime instances) in self which have sufficent
            data to determine this value.

        Examples
        --------
        If there is not enough information to calculate a time ratio, None is
        returned.

        >>> lt_collector = LocationTimeCollector(camera_placements={
        ...     "foo": LocationTime(camera_start_time="2021-07-15T13:00"),
        ... })
        >>> print(lt_collector.get_time_ratio())
        None
        """
        time_ratios: list[float] = []
        for lt in self.camera_placements.values():
            time_ratio = lt.get_time_ratio()
            if time_ratio is not None:
                time_ratios.append(time_ratio)

        if len(time_ratios) == 0:
            return None
        return fsum(time_ratios) / len(time_ratios)

    def get_correctors(
        self, camera_time_to_actual_time_ratio: Optional[float] = None
    ) -> SubDirDict[DatetimeCorrector]:
        """Calls the .corrector method on each LocationTime in self.camera_placements to
        produce a SubDirDict of datetime_corrector functions.

        Parameters
        ----------
        camera_time_to_actual_time_ratio : Optional[float]
            The amount of time elapsed as reported by the camera divided by the actual
            amount of time elapsed. If None, then this is inferred by calling
            self.get_time_ratio(). If this returns None, a RuntimeError is raised.

        Returns
        -------
        datetime_correctors : SubDirDict[DatetimeCorrector]
            A mapping from subdirectories to DatetimeCorrector functions.

        Examples
        --------
        >>> lt_collector = LocationTimeCollector(camera_placements={
        ...     "data": LocationTime(
        ...         camera_start_time="2021-07-15T14:00",
        ...         camera_end_time="2021-07-15T16:00",
        ...         actual_end_time="2021-07-15T15:00",
        ...     ),
        ...     "foo": LocationTime(camera_start_time="2021-07-15T13:00"),
        ... })
        >>> datetime_correctors = lt_collector.get_correctors(
        ...     camera_time_to_actual_time_ratio=2.0
        ... )
        >>> datetime_correctors["data"](datetime(2021, 7, 15, 18, 0))
        datetime.datetime(2021, 7, 15, 16, 0)

        If camera_time_to_actual_time_ratio is not set, it is calculated from
        the camera_placements in the LocationTimeCollector

        >>> datetime_correctors = lt_collector.get_correctors()
        >>> datetime_correctors["data"](datetime(2021, 7, 15, 18, 0))
        datetime.datetime(2021, 7, 15, 16, 0)

        But don't do this if there isn't enough information

        >>> lt_collector = LocationTimeCollector(camera_placements={
        ...     "foo": LocationTime(camera_start_time="2021-07-15T13:00"),
        ... })
        >>> print(lt_collector.get_time_ratio())
        None
        >>> datetime_correctors = lt_collector.get_correctors()
        Traceback (most recent call last):
        ...
        RuntimeError: Unable to calculate camera-to-actual time ratio from time data
        """
        if camera_time_to_actual_time_ratio is None:
            camera_time_to_actual_time_ratio = self.get_time_ratio()

        if camera_time_to_actual_time_ratio is None:
            raise RuntimeError(
                "Unable to calculate camera-to-actual time ratio from time data"
            )

        datetime_correctors: SubDirDict[DatetimeCorrector] = SubDirDict()
        for directory, location_time in self.camera_placements.items():
            datetime_correctors[Path(directory)] = location_time.corrector(
                camera_time_to_actual_time_ratio
            )

        return datetime_correctors

    def get_location_dict(self) -> SubDirDict[Optional[str]]:
        """Produces a SubDirDict of location strings.

        Returns
        -------
        location_dict : SubDirDict
            A mapping from subdirectories to location strings.

        Examples
        --------
        >>> lt_collector = LocationTimeCollector(camera_placements={
        ...     "data": LocationTime(
        ...         camera_start_time="2021-07-15T14:00",
        ...         camera_end_time="2021-07-15T16:00",
        ...         actual_end_time="2021-07-15T15:00",
        ...         location="loc0",
        ...     ),
        ...     "foo": LocationTime(
        ...         camera_start_time="2021-07-15T13:00",
        ...         location="loc1",
        ...     ),
        ...     "bar": LocationTime(camera_start_time="2021-07-15T13:00"),
        ... })
        >>> lt_collector.get_location_dict()
        SubDirDict({Path('data'): 'loc0', Path('foo'): 'loc1', Path('bar'): None})
        """
        location_dict: SubDirDict[Optional[str]] = SubDirDict()
        for directory, location_time in self.camera_placements.items():
            location_dict[Path(directory)] = location_time.location

        return location_dict
